#ifdef IGNORE_ME
  #define ident "@(#) c_array.h (/sccs/ri_util/include/s.c_array.h) 1.3 02/24/99 14:20:44";
#endif

/*  
  Copyright 1997 Kyle R. Burton

This is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

This file is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with log_daemon; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

=head1 NAME

c_array - dynamic array object.

=head1 SYNOPSIS

  #include <c_array.h>
  ...
  c_array<int> numbers;
  numbers[3] = 0;
  numbers[2] = 1;
  numbers[1] = 2;
  numbers[0] = 3;

  cout << "the second element is: " << numbers[1] << endl;

=head1 DESCRIPTION

This is a dynamicly growable array object for C++.  It is tempalte based,
so the elements can take on any class that supports default, and copy, 
constructors, the assignment, and equality operators.

=head1 API REFERENCE

=head2 c_array( int chunk = 0, int initial = 0 )

This is the default constructor.  The two arguments are the chunk value,
and the inital segment.  The chunk value determines how memory is allocated
when the object needs to enlarge itself - memory is allocated in blocks the 
size of the template object times the chunk size.  The initial parameter
determines the initial allocation, and doens't need to be the same size as,
or a multiple of the chunk size.

=head2 c_array( const c_array<T>& that )

This is the copy constructor for c_array.  It takes a const reference to
another array, and makes itself a copy of the other array.

=head2 virtual ~c_array()

The destructor.  Any memory allocated by the c_array object is freed
upon deletion.

=head2 virtual const c_array<T>& operator=( const c_array<T>& right )

This is the assignment operator, you can use it to set one c_array
object equal to another.

=head2 virtual int set_at( int pos, const T& elm )

This is a synonym for operator[].  It allows you to set an 
element at a given index.  Unlike operator[], this function
returns an integer, so you can't use it as an lvalue, or as an rvalue.

=head2 virtual int get_at( int pos, T& elm )

This allows you to retreive an element at a given index from
the c_array object.

=head2 virtual T& operator[]( int pos )

The index operator allows you to set or retreive elements from the
c_array.  It returns a referenece directly to the element in the 
array, this allows you to use it as both an lvalue and an rvalue.

=head2 virtual int find( const T& fnd )

This function allows you to find the index of an element
in the array.  T::operator== (the equality operator) is invoked
for each element in the array, starting at the first, and continuing
untill a match is found, or the end of the array is reached.

=head2 virtual int slice( c_array<T>& ary, int start, int len = -1 )

This function was inspired by perl's slice function.  What it does
is take an array reference (to hold the subsequent array slice),
a starting position, and a length.  It sets ary equal to all 
the elements of the c_array object that it is invoked on, starting
at the given start position, for len elements.  If len is negative,
or if it reaches past the end of the c_array object, only the
elements upto and including the last one are copied.

=head2 virtual int insert( int before, const T& elm )

This function will insert elm into the position just
in front of the element at the position specified by the before
argument.  A value of 0 for before causes the element to be
inserted at the beginning of the array.

=head2 virtual int remove( const T& elm )

remove() invokes find() to locate the first element that matches
elm, and removes it from the array.

=head2 virtual const c_array<T>& operator+=( const T& elm )

'Plus Equals' this function allows you to append, or concatenate,
an element onto the end of the array.

=head2 virtual const c_array<T>& operator+=( const c_array<T>& right )

This versino of the 'Plus Equals' function allows you to append, or concatenate,
an entire array onto the end of the c_array being used as the lvalue.

=head2 friend c_array<T> operator+( const c_array<T>& left, const T& elm )

This function allows an element to be added to the end of an array.  What is produced
is a temporary array which only lives to the end of the current statement.
This function is typicly used as an rvalue because of this.

=head2 friend c_array<T> operator+( const c_array<T>& left, const c_array<T>& right )

This function allows two arrays to be added to each other.  What is produced
is a temporary array which only lives to the end of the current statement.
This function is typicly used as an rvalue because of this.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

L<c_collection>

=cut

*/



////////////////////////////////////////////////////////////////////////////////
// file: c_array.h
//
// description:  declaration, and implementation of template based c_array 
//    object
//
// creatd:  Sat 10/25/1997 14:34:17.26 Kyle R. Burton
//
#include "c_collection.h"
#include <stdio.h>

#if !defined( _C_ARRAY_H )
#define _C_ARRAY_H

#define C_ARRAY_DEF_CHUNK  10

class c_array_exception
  : virtual public c_exception
{
private:
  c_array_exception( const char* msg = (const char*)NULL, char *file = (char*)NULL, int line = 0 ) 
    : c_exception( msg, file, line ) {
  };
public:
  ~c_array_exception() {
  };

  static c_array_exception* New( const char* msg, char *file, int line ) {
    return( new c_array_exception( msg, file, line ) );
  };
};


//
// class:  c_array
//
// description:  dynamicly growable template based array object
//
template<class T>
class c_array 
  : virtual public c_collection<T>
{
public:
  //
  // default constructor
  //  chunk  - internal array will be grown by chunk during reallocations
  //  initial  - the initial extent of the array
  //
  c_array( int chunk = 0, int initial = 0 ) {
    init();
    m_chunk = (chunk > 0) ? chunk : C_ARRAY_DEF_CHUNK;
    re_alloc( initial > 0 ? initial : C_ARRAY_DEF_CHUNK );
  };

  //
  // copy constructor
  //
  c_array( const c_array<T>& that ) {
    init();
    *this = that;
  };

  //
  // destructor
  //
  virtual ~c_array() {
    delete [] m_array;
  };

  // implementation of c_collection interface
  virtual int add( const T& elm ) {
    (*this)[m_elms] = elm;
    return( TRUE );
  };

  virtual int remove_all( void ) {
    for( int i = 0; i < m_elms; i++ ) {
      m_array[i] = T();
    }
    m_elms = 0;

    return( TRUE );
  };

  // virtual c_collection_iter<T>* prepare_iter( const collection_sorting_order ord = raw_order ) {
  virtual c_collection_iter<T>* prepare_iter( const collection_sorting_order ord = raw_order ) const {
    return( new c_collection_iter<T>( c_collection<T>::owner_id(this), NULL, ord ) );
  };

  virtual int first( c_collection_iter<T>* iter, T& elm ) const { 
    if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    if( NULL == m_array || m_elms <= 0 ) {
      return( FALSE );
    }

    switch( iter->m_ord ) {
      case pre_order:
      case raw_order:
        iter->m_data = m_array;
        elm = *((T*)iter->m_data);
        break;
      case post_order:
        iter->m_data = &(m_array[m_elms-1]);
        elm = *((T*)iter->m_data);
        break;
      default:
        throw( c_array_exception::New( "Invalid sorting order", __FILE__, __LINE__ ) );
        // break;
    }

    return( TRUE );
  };
  
  virtual int next( c_collection_iter<T>* iter, T& elm ) const { 
    T* p;
    if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    if( m_elms <= 0 ) {
      return( FALSE );
    }

    switch( iter->m_ord ) {
      case raw_order:
      case pre_order:
        if( NULL == iter->m_data ) {
          return( first( iter, elm ) );
        }

        if( iter->m_data >= &(m_array[m_elms-1]) ) {
          return( FALSE );
        }

        p = (T*)iter->m_data;
        p++;
        iter->m_data = p;
        elm = *p;
        break;
      case post_order:
        if( NULL == iter->m_data ) {
          return( last( iter, elm ) );
        }

        if( iter->m_data <= m_array ) {
          return( FALSE );
        }
        p = (T*)iter->m_data;
        p--;
        iter->m_data = p;
        elm = *p;
        break;
      default:
        throw( c_array_exception::New( "Invalid sorting order", __FILE__, __LINE__ ) );
        // break;
    }

    return( TRUE );
  };
  
  virtual int prev( c_collection_iter<T>* iter, T& elm ) const { 
    T *p;
    if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    if( m_elms <= 0 ) {
      return( FALSE );
    }

    switch( iter->m_ord ) {
      case raw_order:
      case pre_order:
        if( NULL == iter->m_data ) {
          return( last( iter, elm ) );
        }

        if( iter->m_data <= m_array ) {
          return( FALSE );
        }

        p = (T*)iter->m_data;
        p--;
        iter->m_data = p;
        elm = *p;
        break;
      case post_order:
        if( NULL == iter->m_data ) {
          return( first( iter, elm ) );
        }

        if( iter->m_data >= &(m_array[m_elms-1]) ) {
          return( FALSE );
        }
        p = (T*)iter->m_data;
        p++;
        iter->m_data = p;
        elm = *p;
        break;
      default:
        throw( c_array_exception::New( "Invalid sorting order", __FILE__, __LINE__ ) );
        // break;
    }

    return( TRUE );
  };

  virtual int last( c_collection_iter<T>* iter, T& elm ) const { 
    if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    if( NULL == m_array || m_elms <= 0) {
      return( FALSE );
    }

    switch( iter->m_ord ) {
      case raw_order:
      case pre_order:
        iter->m_data = &(m_array[m_elms-1]);
        elm = *((T*)iter->m_data);
        break;
      case post_order:
        iter->m_data = &(m_array[0]);
        elm = *((T*)iter->m_data);
        break;
      default:
        throw( c_array_exception::New( "invalid sort order", __FILE__, __LINE__ ) );
        // break;
    }

    return( TRUE );
  };

  virtual int complete_iter( c_collection_iter<T>* iter ) const {
    delete iter;
    return( TRUE );
  };


  //
  // assingment operator
  //
  virtual const c_array<T>& operator=( const c_array<T>& right ) {
    if( m_size < right.m_size ) {
      re_alloc( right.m_size );
    }

    m_chunk = right.m_chunk;
    m_elms = right.m_elms;

    for( int i = 0; i < m_elms; i++ ) {
      m_array[i] = right.m_array[i];
    }

    return( *this );
  };

  //
  // sets the element at pos to elm, returns TRUE on success, FALSE
  // on failure
  //
  virtual int set_at( int pos, const T& elm ) {
    int rv = TRUE;
    if( pos < 0 || pos >= m_elms ) {
      rv = FALSE;
    }
    else {
      m_array[pos] = elm;
    }

    return( rv );
  };

  //
  // returns refrence to the element at pos, throws c_array_exception if pos
  // is an invalid array index (-1, or grater than the number of elemnts in 
  // the array).
  // 
  virtual T& operator[]( int pos ) {
    if( pos < 0 ) {
      throw( c_array_exception::New( "negative array index", __FILE__, __LINE__ ) );
    }

    if( pos >= m_size ) {
      re_alloc( pos + m_chunk );
    }

    // pos is zero based, m_elms is ones based...
    if( (pos+1) > m_elms ) {
      m_elms = pos + 1;
    }

    return( m_array[pos] );
  };

  //
  // sets elm to the element at pos, returns TRUE on success,
  // FALSE on failure
  //
  virtual int get_at( int pos, T& elm ) {
    int rv = TRUE;
    if( pos < 0 ) {
      rv = FALSE;
    }
    else {
      if( pos >= m_size ) {
        re_alloc( pos + m_chunk );
      }

      if( (pos+1) > m_elms ) {
        m_elms = pos + 1;
      }

      elm = m_array[pos];
    }

    return( rv );
  };

  //
  // returns the number of elements in the array
  //
  virtual int get_count( void ) const {
    return( m_elms );
  };

  // 
  // returns index of element that matches fnd, or -1 if not found
  //
  virtual int find( const T& fnd ) {
    int pos = -1, i;
    T tmp;

    for( i = 0; i < m_elms; i++ ) {
      get_at( i, tmp );
      if( fnd == tmp ) {
        pos = i;
        break;
      }
    }

    return( pos );
  };

  virtual int slice( c_array<T>& ary, int start, int len = -1 ) {
    T elm;

    ary.remove_all();

    start = (start < 0) ? 0 : start;
    if( len < 0 || (len+start) > m_elms ) {
      len = m_elms;
    }

    for( int i = 0; i < len; i++ ) {
      get_at( i + start, elm );
      ary[i] = elm;
    }

    return( TRUE );
  };

  virtual int insert( int before, const T& elm ) {
    c_array<T> first_half, second_half;

    before = (before < 1) ? 1 : (before >= m_elms) ? m_elms -1 : before;

    slice( first_half, 0, before - 1 );
    slice( second_half, before - 1 );
    first_half += elm;
    *this = first_half + second_half;

    return( TRUE );
  };

  virtual int remove( const T& elm ) {
    int pos = find( elm );

    if( -1 == pos ) {
      return( FALSE );
    }

    c_array<T> first_half, second_half;
    slice( first_half, 0, pos - 1 );
    slice( second_half, pos );
    *this = first_half + second_half;

    return( TRUE );
  };

  virtual const c_array<T>& operator+=( const T& elm ) {
    add( elm );
    return( *this );
  };

  virtual const c_array<T>& operator+=( const c_array<T>& right )
  {
    c_collection_iter<T> *iter = right.prepare_iter( pre_order );
    T elm;

    while( right.next( iter, elm ) ) {
      add( elm );
    }

    right.complete_iter( iter );

    return( *this );
  };

  friend c_array<T> operator+( const c_array<T>& left, const T& elm ) {
    c_array<T> ary( left );
    ary.add(elm);
    return( ary );
  };

  friend c_array<T> operator+( const c_array<T>& left, const c_array<T>& right ) {
    c_array<T> ary( left );
    ary += right;
    return( ary );
  };

private:
  T*  m_array;  // the actual array
  int m_chunk,  // the chunk size
    m_size,    // the allocated size
    m_elms;    // the number of elements used (out of m_size)

  void init( void ) {
    m_array = (T*)NULL;
    m_chunk = C_ARRAY_DEF_CHUNK;
    m_size = 0;
    m_elms = 0;
  };

  void re_alloc( int size ) {
    if( size > m_size ) {
      T *p_new = new T[size];
      // copy the elements from old to new...
      for( int i = 0; i < m_size; i++ ) {
        p_new[i] = m_array[i];
      }

      delete [] m_array;
      m_array = p_new;
      m_size = size;
    }
  };
};

#endif // !defined( _C_ARRAY_H )
