#ifdef IGNORE_ME
  #define ident "@(#) c_auto_ptr.h (/sccs/ri_util/include/s.c_auto_ptr.h) 1.3 02/24/99 14:20:44";
#endif

/*  c_auto_ptr.h
  Copyright 1998 Kyle R. Burton

c_auto_ptr.h is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

c_auto_ptr.h is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with c_auto_ptr.h; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

=head1 NAME c_auto_ptr

Automatic memory allocation and cleanup.

=head1 SYNOPSIS

  #include <c_auto_ptr.h>
  ...
  {
    c_auto_ptr<char*> buff(1024);
    sprintf(buff,"this is a string of text\n");
    cout << "The string was: " << buff.get_ptr() << endl;
    throw(1);
  }

=head1 DESCRIPTION

The auto pointer object can be used as a 'set and forget' memory
allocator.  It can be used in contexts where a template type
pointer is expected, and it will autocast itself to the pointer type.
It is convienient because if the scope of execution leaves the scope
where the object was constructed, it cleans up the allocated memory -
this is especially useful during the stack unwind that ocurrs during
a thrown exception.  Under exception handeling, it is often difficult 
to program around the issue of re-entering a block of code, or allowing
for the catching code to free memory allocated for non-autos.

=head1 API REFERENCE

=head2 c_auto_ptr(void)

The default constructor.  It sets aside a size of 1 for the template
type.

=head2 c_auto_ptr( unsigned int _size )

This constructor takes an unsigned integer to use to size
the allocated memory.  It will allocate _size * sizeof(T)
bytes of memory, enough to hold _size objects.

=head2 ~c_auto_ptr(void)

The destructor - this frees any memory allocated by the c_auto_pointer
object via the delete [] operator.

=head2 T* get_ptr(void)

This function returns a pointer to the beginning of the internal
buffer.

T* operator&(void)

This function returns a pointer to the beginning of the internal
buffer.

=head2 unsigned int size(void)

This returns the size of the internal buffer that was
originaly allocated -- this is not in bytes, but it is 
instead bytes/sizeof(T).

=head2 operator T*( void ) const

This provides autocasting to (const*) for the template type (T).
Thus if a function or assignment expects a const T*, this function
will autocast the c_auto_pointer object to that type.

=head2 int realloc( unsigned int size )

This function will reallocate the internal buffer using the
C function call realloc().

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=cut

*/

#if !defined( _C_AUTO_PTR_H )
#define _C_AUTO_PTR_H 

#include <stdinc.h>

/*
 * class c_auto_ptr
 *
 *
 *
 */
template<class T>
class c_auto_ptr {
public:
  c_auto_ptr(void )
    : buff_size(1) 
    , data((T*)NULL) {
    realloc( buff_size );
  };

  c_auto_ptr( unsigned int _size )
    : buff_size(_size)
    , data((T*)NULL) {
    realloc(buff_size);
  };

  ~c_auto_ptr(void ) {
    delete [] data;
  };

  T* get_ptr(void) {
    return data;
  };

  T* operator&( void ) {
    return get_ptr();
  };

  unsigned int size(void) {
    return buff_size;
  };

  operator T*( void ) const { return data; };

  int realloc( unsigned int size ) {
    delete [] data;
    buff_size = size;
    data = new T[buff_size];
    return buff_size;
  };

private:
  T*  data;
  unsigned int buff_size;
};

#endif // !defined( _C_AUTO_PTR_H )
