#ifdef IGNORE_ME
  #define ident "@(#) c_collection.h (/sccs/ri_util/include/s.c_collection.h) 1.7 02/24/99 14:20:44";
#endif

  /*  c_collection.h
  Copyright 1997 Kyle R. Burton

This file is part of c_collection.h, an attempt at a win32 equivalent for
the system log daemon available under unix.

c_collection.h is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

c_collection.h is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with c_collection.h; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

=head1 NAME

c_collection - base class for collection objects

=head1 SYNOPSIS

  template class<T>
  void dump_elements( const c_collection<T>& col )
  {
    T tmp;
    c_collection_iter<int> *iter = col.prepare_iter();
    while(col.next(iter,tmp)) {
      cout << "next item is:  " << tmp << endl;
    }
    col.complete_iter(iter);
  }

=head1 DESCRIPTION

This is the base class for the collection objects. 
It is a virtual base class, and you therefor you 
cannot create one directly.

=head1 API REFERENCE

=head2 c_collection()

This is the default constructor.  It does nothing.

=head2 virtual int add( const T& elm )

This appends a new element to the end of the array, effectivly growing the
array.  It may or may not cause a memory allocation depending on how much
of the currently allocated chunk is used.

=head2 c_collection( const c_collection& that )

Copy constructor.

=head2 virtual ~c_collection()

Destructor.

=head2 virtual int get_count( void ) const

Returns a count of the number of elements in the collection.
This must be implemented by the derived class.  This function
will not modify the collection object, or it's derived object.

=head2 virtual int add( const T& elm ) 

Used to add an element to a collection.  The element is typicly added
through T::T(const T&), though it may be added through T::operator=(const T&).
The behavior of this member function is entirely dependant on the derived
class.  The typical action of add() is to append an element to the collection 
(though 'append' may mean different things for different dervied classes).

=head2 virtual int remove_all( void ) 

This is used to clear out the collection.  It must be implemented by the
derived class.

=head2 virtual c_collection_iter<T>* prepare_iter( const collection_sorting_order ord = raw_order ) const

This is the first member function in the iteration interface.  It is used
to create an iterator that can then be used to fetch each of the elements
from the collection.  The ordering of the elements when they are fetched 
is entirely determined by the derived class.

Please be aware that since the collection iterator objects are seporate
from the collection class that creates them, they store state object
associated with the iteration that is being taken place.  If you add
or remove elements from a collection during an iteration, the resulting
behavior of the system will be undefiend.

=head2 virtual int first( c_collection_iter<T>*, T& ) const

=head2 virtual int next( c_collection_iter<T>*, T& ) const

=head2 virtual int prev( c_collection_iter<T>*, T& ) const

=head2 virtual int last( c_collection_iter<T>*, T& ) const

These functions are used to retreive, and iterate within the collection
class.  It is not necessary to call first() or last() before calling 
next() or prev() - next() and prev() will act as first() or last() 
respectivly if they are the first iteration functinons invoked.
These functions do not modify the collection object.

=head2 virtual int complete_iter( c_collection_iter<T>* iter ) const

Call this function to complete an iteration.  This is necessary because
the collection stores some of it's state within the iterator pointer, so
it is necessary to allow the collection object to clean up after itself.
This function does not modify the collection object.

=BUGS

No known bugs.

=AUTHORS

Kyle R. Burton mortis@voicenet.com

=SEE ALSO

L<c_stack>
L<c_list>
L<c_array>
L<c_hash>

=cut

*/


#include "c_exception.h"

#if !defined( _C_COLLECTION_H )
#define _C_COLLECTION_H

class c_collection_exception 
  : virtual public c_exception
{
  public:

  c_collection_exception( const char* msg, char* file, int line ) 
    : c_exception( msg, file, line ) {
    ;// do nothing
  };

  ~c_collection_exception() {
    ;// do nothing
  };

  // virtual c_collection_exception* New( const char* msg, char* file, int line ) = 0;

  virtual const c_exception& operator=( const c_exception& right ) {
    c_exception::operator=( right );
    return *this;
  };

  virtual const c_collection_exception& operator=( const c_collection_exception& right ) {
    c_exception::operator=( right );
    return *this;
  };
};

template<class T> class c_collection_iter;

// raw_order logicly is "as is" as the collection stores it (i.e. no order)
// pre_order - usually sorted ascending - but as implemented by the derived collection
// post_order - usually sorted descending - but as implemented by the derived collection
enum collection_sorting_order { raw_order = 0, pre_order, post_order };

//
// class:  c_collection
//
// description:  pure virtual base class for collections.  Whenever
//    possible, code should use the c_collection_iter, and 
//    c_collection interfaces, when the extended functionality of
//    the specific collections isn't needed.
//
typedef void* owner_id;
template<class T> class c_collection 
{
public:

  c_collection() {
  };

  c_collection( const c_collection& that ) {
    *this = that;
  };

  virtual ~c_collection() {
    ;// do nothing
  };

  virtual int get_count( void ) const = 0;

  // add an element - simplest implementation possible for your
  // derived collection class, this should be a no-brainer for
  // stack, list, and array objects, but it will need to be 
  // implemented at your discretion for map, and hash type 
  // objects...
  virtual int add( const T& elm ) = 0;

  // delete all the elements in the collection
  virtual int remove_all( void ) = 0;

  // get an iterator, use this function to allocate an iterator object
  virtual c_collection_iter<T>* prepare_iter( const collection_sorting_order ord = raw_order ) const = 0;

  // useing the iterator from the return of the above...
  // get the first element
  virtual int first( c_collection_iter<T>*, T& ) const { return( FALSE ); };
  // get the next element
  virtual int next( c_collection_iter<T>*, T& ) const { return( FALSE ); };
  // get the previous element
  virtual int prev( c_collection_iter<T>*, T& ) const { return( FALSE ); };
  // get the last element
  virtual int last( c_collection_iter<T>*, T& ) const { return( FALSE ); };

  // use this function to de-allocate the iterator object
  virtual int complete_iter( c_collection_iter<T>* iter ) const {
    delete iter;
    return( TRUE );
  };
};

//
// class:  c_collection_iter
//
// description:  pure virtual base class, object used to iterate
//    through the derived collection classes.  
//
template<class T>
class c_collection_iter
{
public:
  c_collection_iter( 
    // c_collection<T>::owner_id owner_id = 0,
    void* owner_id = NULL,
    void* data = NULL,
    collection_sorting_order ord = raw_order 
  ) 
    : m_ord( ord )
    , m_data( data )
    , m_owner_id( owner_id ) {
  };

  c_collection_iter( const c_collection_iter& that ) 
    : m_ord( that.m_ord )
    , m_data( that.m_data )
    , m_owner_id( that.m_owner_id ) {
  };

  virtual ~c_collection_iter() {
    ;// do nothing
  };

  virtual const c_collection_iter& operator=( const c_collection_iter& right ) {
    m_ord = right.m_ord;
    m_data = right.m_data;
    m_owner_id = right.m_owner_id;

    return(*this);
  };


  virtual void set_sort( collection_sorting_order ord ) {
    m_ord = ord;
  };

public:
  // the sorting order
  collection_sorting_order  m_ord;

  // whatever the base collection class needs to store here...
  void*          m_data;
  // c_collection<T>::owner_id  m_owner_id;      
  owner_id      m_owner_id;      
};


#endif // !defined( _C_COLLECTION_H )
