#ifdef IGNORE_ME
  #define ident "@(#) c_dump_tree_to_files.h (/sccs/ri_util/include/s.c_dump_tree_to_files.h) 1.5 02/24/99 14:20:45";
#endif

/*
 * file: c_dump_tree_to_files.h
 *
 * description:  this object will take a multi node tree that logicly
 *      represents database records, and break it out into delimited files.
 *
 * revision list:
 *   Wed Aug 26 17:30:24 EDT 1998 KRB - added the header info...
 *
 */

/*

=head1 NAME 

c_dump_tree_to_file - dumps a tree into files

=head1 SYNOPSIS

  #include <c_dump_tree_to_files.h>
  ...
  class c_string_value_to_string: virtual public c_value_to_string<c_string> {
  public:
    virtual int to_string(const c_string& key, c_string &s) {
      s = key + ".csv"; return TRUE;
    };
  };
  ...
  c_string_value_to_string convertor;
  c_dump_tree_to_file<c_string,c_string,char,char> dumper(conv,';','\n');
  dumper.dump(handle);

=head1 DESCRIPTION

This object is for taking a c_multi_tree, and dumping it's values to
a series of files.  It decides how to perform the dump by examining
the nodes in the tree.  If a node has children, it is consdiered to be 
a sub-tree, and it calls c_value_to_string::to_string() on that node's
key to produce a file name.  Any leaf nodes that are a direct child
of that sub-tree node are then dumped into that file.  A leaf node is
a node that has no child nodes.  

This is tpyicly used to dump a tree that models a heirarchical collection
of data, such as a grouping of related records from an RDBMS, and then
dump them into seporate files.  Based on tables for example.

=head1 API REFERENCE

=head2 c_dump_tree_to_file( )

This is the default constructor.  It takes three arguments, and all of
them are required.  

The first is a c_value_to_string derived object.
This object must know how to calculate a file name from a key element
from the tree.  

The second parameter is a field delimiter.  
A field delimiter is inserted inbetween leaf nodes that are dumped to 
the same file.  The field delimiter object must support streaming via 
the fstream objects.

The third is a record delmimiter.  This also is streamed to an fstream
object, and is appended after all of the leaf nodes for a given sub-tree
in the tree have been output.

=head2 virtual ~c_dump_tree_to_file( void )

This is the destructor.   

=head2 virtual int dump( c_multi_tree_manipulator<KEY,VALUE>& handle )

This is the function that invokes the dumping action.  It takes a 
tree manipulator that should be pre-positioned at the element 
the user wishes the dump to start from.  It then dumps using
the logic described above.

=head2 virtual int dump( c_multi_tree<KEY,VALUE>& tree )

This function performs the same action as the dump() function
that takes the manipulator, only this takes the tree object, and
constructs it's own manipulator object.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

L<c_multi_tree>

=cut

*/

#if !defined( _C_DUMP_TREE_TO_FILE_H )
#define  _C_DUMP_TREE_TO_FILE_H

#include <iostream.h>
#include <fstream.h>

#include <stdinc.h>
#include <c_timestamp.h>
#include <c_string.h>
#include <c_stack.h>
#include <c_hash.h>
#include <c_multi_tree.h>


template<class KEY>
class c_value_to_string {
public:
  c_value_to_string( void ) {
    ;// do nothing...
  };

  virtual ~c_value_to_string( void ) {
    ;// do nothing
  };

  // this is what you have to override...
  virtual int to_string( const KEY& key, c_string& s ) = 0;
private:
};

template<
  class KEY,
  class VALUE,
  class FIELD_DELIMITER,
  class RECORD_DELIMITER
>
class c_dump_tree_to_file {
public:
  c_dump_tree_to_file( 
      c_value_to_string<KEY>& k_to_s, 
      const FIELD_DELIMITER& fd, 
      const RECORD_DELIMITER& rd 
  ) : field_sep(fd)
    , record_sep(rd)
    , value_to_string(k_to_s) {
    ;// do nothing...
  };

  virtual ~c_dump_tree_to_file( void ) {
    // close off and release all the ofstreams...
    c_collection_iter<c_map_pair<c_string,ofstream*> > *iter = streams.prepare_iter();
    c_map_pair<c_string,ofstream*> pair;
    while( streams.next(iter,pair) ) {
      pair.m_value->close();
      delete pair.m_value;
    }
    streams.complete_iter(iter);
  };

  // traverse the tree from where handle points at it, and dump
  // it out to files...
  virtual int dump( c_multi_tree_manipulator<KEY,VALUE>& handle ) {
    c_multi_tree_manipulator<KEY,VALUE> _handle(handle);
    c_stack<c_string> files;
    return _dump(_handle,files);
  };

  // overload to take a tree instead of a handle...
  virtual int dump( c_multi_tree<KEY,VALUE>& tree ) {
    c_multi_tree_manipulator<KEY,VALUE> _handle(tree);
    c_stack<c_string> files;
    return _dump(_handle,files);
  };

private:
  // this is the recursive one...and you can't call it from the outside...
  virtual int _dump( 
    c_multi_tree_manipulator<KEY,VALUE>& handle,
    c_stack<c_string>& files 
  ) {
    int num_children = handle.num_children();

    // is this a leaf node?
    if( num_children < 1 ) {
      // dont' recurse any further...go back...
      return FALSE;
    }
    else {
      // dump the leaves, and recurse into the sub-trees...
      ofstream* pofstream = NULL;
      c_string value, file;

      int leaves = 0;
      for(int i=0;i<num_children;++i) {
        handle.position(i);
        if( handle.childs_num_children() ) {
          // it's a sub tree
          handle.down();
          _dump(handle,files);
          handle.up();
          continue;
        }
        else {
          if( !pofstream ) {
            handle.get_value(value);
            if( !value_to_string.to_string(value,file) ) {
              INFO(cout) << "error translating value to file name?" << endl;
            }
      
            if( !streams.find(file,pofstream) ) {
              // we need to open it...
              pofstream = new ofstream;
              pofstream->open(file,ios::app|ios::out);
              if( pofstream->fail() ) {
                INFO(cout) << "Error opening file: " << file << " for writing..." << endl;
              }
      
              // stick it in the hash table...
              streams.add(file,pofstream);
            }
          }

          // it's just a leaf node...
          // emit a field delimiter if we've emited at least one field...
          if( leaves ) {
            (*pofstream) << field_sep;
          }

          // emit the field data...
          handle.get_child_value(value);
          (*pofstream) << value;
          ++leaves;
        }
      }

      // ok, we've processed all the children of this node,
      // if we emited any fields, emit a record seporator, and
      // return.
      if( leaves ) {
        (*pofstream) << record_sep;
      }
    }

    return TRUE;
  };

private:
  c_hash<c_string,ofstream*>  streams;
  c_value_to_string<KEY>&     value_to_string;
  FIELD_DELIMITER             field_sep;
  RECORD_DELIMITER            record_sep;
};


#endif // !defined( _C_DUMP_TREE_TO_FILE_H )
