#ifdef IGNORE_ME
  #define ident "@(#) c_hash.h (/sccs/ri_util/include/s.c_hash.h) 1.5 02/24/99 14:20:44";
#endif

/*  c_hash.h
    Copyright 1998 Kyle R. Burton

c_hash.h is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

c_hash.h is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with c_hash.h see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

=head1 NAME c_hash - hashtable object for C++

=head1 SYNOPSIS

  #include <c_hash.h>
  #include <c_string.h>
  ...
  c_hash<c_string,c_string> inventors;

  inventors.add(c_string("Perl"),c_string("Larry Wall");
  inventors.add(c_string("Linux"),c_string("Linus Torvalds");
  inventors.add(c_string("Lightbulb"),c_string("Thomas Edison");
  inventors.add(c_string("Emacs"),c_string("Richard Stallman");

  c_string person;
  inventors.get(c_string("Lightblub"),person);
  cout << "The inventor of the Lightbulb was: " << person << endl;


=head1 DESCRIPTION

c_hash was inspired by the usefulness and utility of Perl's built in
hash datatype.  It can be used to store key/value pairs where
key uniqueness is enforced, and where insertion and lookup 
occurr in significantly less than linear time as compared to
the number of elements stored in the table.  It performs it's
own memory management, and keeps copies of the data that
you attempt to store in it -- not the data itself.  This trades
a bit of performance (i.e. more use of the copy constructor,
and assignmnet operator) for safety, robustness, and encapsualtion.
You can't get at the data in side the object, and it promises to
manage it for you.

c_hash is a template based object, and can therefore store
any object type as a key and a value.  There are restrictions
placed on the objects you are going to store in the table.

  The object must support a default constructor    T::T()
  The object must support a copy constructor       T::T(const T&)
  The object must support an assignmnet operator   const T& T::operator=(const T&)
  The object must support an equality operator     int T::operator==(const T&)

Additionaly, you must provide a function with the following
prototype for the object you are going to use as the key:

  hash_value hash( int m, object O )

Where hash_value is an unsigned long, and m is the base table
size as an integer.  Your function should return a hash_value
in the range 0 <= num < m.  This is then used as an index into
the table for insertion and retreival.  Please take care when
creating this function so the results are as evenly distributed
as possible, or you will advesly effect the performance of the
hash object.  If you know enough detail about your key set, it
would be possible to implement your own hash() function with 
optimization for the key set.

You may wish to make your hash() function in-line to increace 
performance.  A function is already provided for c_string objects.  

A note of warning to those not well versed in the languages 
of C and C++.  Don't use char* as a key and expect to get the 
classical behavior of a hashtable (such as the behavior you 
get with c_string).  Use of a pointer type as the key is 
alsmost gauranteed to give you unintended behavior.  Buy 
using a pointer, you are effectivly using a unique key for 
each and every element you attempt to insert into the 
hashtable.  Since the c_hash object manages it's own memory, 
you will end up incorrectly placing a copy of your pointer in 
the c_hash object.

=head1 API REFERENCE

=head2 c_hash( int size_hint = C_HASH_DEF_HINT ) 

This is the default constructor, it creates the initial hash table.
The parameter size_hint is used to fix the size of the hashtable object.
If you know approximatly how many elements you will be storing in the
hashtable, you can use this setting to pre-tune (to a point) both the
memory, and processing performance for the object.  A larger number 
allows for more elements.

This parameter determines the 'm' value that is passed to the hash()
function described elsewhere in this document.

Unless you know explicit details about the hash() function for your 
object, and really know what you're doing, it's probably a good idea 
to set this to a prime number.  The default is 137.  

=head2 c_hash( const c_hash<K,V>& that ) 

This is a copy cosntructor for the hash object.  It will create a copy
of the other hash table object.

=head2 virtual ~c_hash( void )

This is the destructor for the hash object.  It will clean up all
memory allocated by the hash object.

=head2 virtual int find( K& key ) const 

This is used to test to see if a given key is present in the 
hashtable.  find() returns true if the key is present, or false
if it is not.

=head2 virtual int find( const K& key, V& value ) const

This version of find() can be used to retreive a specified key's 
value from the table.  It returns true on success, and false on 
failure.  Operator V::operator== is invoked to set value to the 
value associated with the specified key.

=head2 virtual int find( c_map_pair<K,V>& pair ) const

This version is simialr to the previou version with the 
exception that it uses the c_map_pair object.  This object
is used because the hash object internaly uses c_map objects 
as the nodes in the table.  It will look for the key element
of the pair, and set the value element of the pair if found.
It reutrns true on success, and false on failure.

=head2 virtual int exists( const K& key ) const

This is a synonym for the first version of find().  It
returns true if the given key is in the hashtable, and 
false otherwise.

=head2 virtual int add( const K& key )

This adds a new pair to the table, or blanks out the value of an
existing pair in the table.  Returns true on success, and false on
failure.  Both V::V(), and V::operator= are invoked during this
function call.

=head2 virtual int add( const K& key, const V& value )

This version of add() takes the value that will be set in the
table.  It invokes V::operator= to set the value.  True is
returned on success, otherwise false is returned.

=head2 virtual int add( const c_map_pair<K,V>& pair )

This is identical to the previous version, with the exception
that it takes a c_map_pair object.

=head2 virtual int clear_values( void )

When invoked on a hash object, the object will walk it's own
hash table, and set the value of each and every pair to V().

=head2 virtual const c_hash<K,V>& operator=( const c_hash<K,V>& right )

This is the assingment operator.  It makes the current hash object
a copy of the other.

=head2 virtual int operator==( const c_hash<K,V>& right )

Equality operator.  This tests weather or not two hash tables
are equal -- meaning that they have the same key/value pairs.
Where each key and each value appears in each hash table.

=head2 virtual int remove( const K& key )

This function can be used to remove a specific element from
the hashtable.  Returns true on success, and false on failure.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

L<c_collection>
L<c_map>

=cut

*/

#if !defined( _C_HASH_H )
#define _C_HASH_H 

#include "c_collection.h"
#include "c_map.h"

typedef unsigned long hash_value;

// this should be a prime number, or you'll end up with holes in the table...
// const int C_HASH_DEF_HINT = 11;
const int C_HASH_DEF_HINT = 137;

const hash_value MAX_HV = (~((hash_value)0));


template<class K, class V>
struct hash_iter {
  hash_value        m_hv;
  c_collection_iter<c_map_pair<K,V> >  *m_iter;
};

template<class K, class V>
class c_hash 
  : virtual public c_collection<c_map_pair<K,V> >
{
public:
  c_hash( int size_hint = C_HASH_DEF_HINT ) 
    : m_table_size( size_hint ) 
    , m_table(0) {
    m_table = new c_map<K,V>[m_table_size];
  };

  c_hash( const c_hash<K,V>& that ) 
    : m_table_size(0)
    , m_table(0) {
    *this = that;
  };

  virtual ~c_hash( void ) {
    delete [] m_table;
  };

  virtual int find( K& key ) const {
    hash_value hv = hash( m_table_size, key );
    return( m_table[hv].find( key ) );
  };

  virtual int find( const K& key, V& value ) const {
    // **b $FileName:$LineNumber
    hash_value hv = hash( m_table_size, key );
    c_map_pair<K,V> pair(key,V());
    if( m_table[hv].find( pair ) ) {
      value = pair.m_value;
      return TRUE;
    }
    return FALSE;
  };

  virtual int find( c_map_pair<K,V>& pair ) const {
    // **b $FileName:$LineNumber
    hash_value hv = hash( m_table_size, pair.m_key );
    return m_table[hv].find(pair);
  }

  virtual int exists( const K& key ) const {
    // _HEY_YOU_: if hash() is causing an error for your KEY
    //  then _YOU_ probably forgot to write a hash(int,const KEY)
    //  for your KEY.  hash(int,const char*) is provided
    //  at the end of this file as an example. (and as a default
    //  for const char*.
    hash_value hv = hash( m_table_size, key );
    return( m_table[hv].is_key(key) );
  };

  virtual int add( const K& key ) {
    hash_value hv = hash( m_table_size, key );
    return(m_table[hv].add(key));
  };

  virtual int add( const K& key, const V& value ) {
    hash_value hv = hash( m_table_size, key );
    #if defined( __DEBUG )
      cout << __FILE__ << "(" << __LINE__ << ") "
           << "adding (" << key << "," << value << ")" << endl;
    #endif
    return( m_table[hv].add( c_map_pair<K,V>(key,value) ) );
  };

  virtual int add( const c_map_pair<K,V>& pair ) {
    hash_value hv = hash( m_table_size, pair.m_key );
    return( m_table[hv].add( pair ) );
  };

  virtual int get_count( void ) const {  // c_collection interface...
    int total = 0;
    for( hash_value i = 0; i < m_table_size; i++ ) {
      total += m_table[i].get_count();
    }
    return( total );
  };

  virtual int remove_all( void ) {
    for( hash_value i = 0; i < m_table_size; i++ ) {
      m_table[i].remove_all();
    }

    return( TRUE );
  };

  virtual int clear_values( void ) {
    for( hash_value i = 0; i < m_table_size; i++ ) {
      m_table[i].clear_values();
    }

    return TRUE;
  };


  virtual c_collection_iter<c_map_pair<K,V> >* prepare_iter( 
    const enum collection_sorting_order ord = raw_order ) const 
  {
    return( new c_collection_iter<c_map_pair<K,V> >( 
      // c_collection<c_map_pair<K,V> >::owner_id(this), 
      owner_id(this), 
      NULL, 
      ord  )
    );
  };

  virtual int first( c_collection_iter<c_map_pair<K,V> > *iter, c_map_pair<K,V>& key ) const {
    hash_iter<K,V> *hi;
    c_map_pair<K,V> pair;
    int rv = FALSE;
    if( iter->m_ord == post_order ) {
      rv = last(iter,key);
    }
    else {
      if ( NULL == iter->m_data ) {
        hi = new hash_iter<K,V>;
        iter->m_data = hi;
        hi->m_hv = 0;
        hi->m_iter = (c_collection_iter<c_map_pair<K,V> >*)NULL;
      }
      else {
        hi = (hash_iter<K,V>*)iter->m_data;
        if( NULL != hi->m_iter ) {
          m_table[hi->m_hv].complete_iter(hi->m_iter);
        }
      }

      hi->m_hv = 0;
      while( hi->m_hv < m_table_size ) {
        hi->m_iter = m_table[hi->m_hv].prepare_iter(iter->m_ord);
        if( m_table[hi->m_hv].first(hi->m_iter,pair) ) {
          rv = TRUE;
          key = pair;
          break;
        }
        m_table[hi->m_hv].complete_iter(hi->m_iter );
        hi->m_hv++;
      }
    }
    return( rv );
  };

  virtual int next( c_collection_iter<c_map_pair<K,V> > *iter, c_map_pair<K,V>& key ) const {
    hash_iter<K,V> *hi = (hash_iter<K,V>*)iter->m_data;
    int rv = FALSE;
    c_map_pair<K,V> pair;

    if( NULL == hi ) {
      rv = first(iter,key);
    }
    else {
      pair = key;
      if( !m_table[hi->m_hv].next(hi->m_iter,pair) ) {
        do {
          // _DEBUG_COND( 
          //  cout << "incrementing hash: " << hi->m_hv << endl 
          // ); 
          m_table[hi->m_hv++].complete_iter(hi->m_iter);
          if( hi->m_hv >= m_table_size ) {
            rv = FALSE;
            break;
          }
          hi->m_iter = m_table[hi->m_hv].prepare_iter(iter->m_ord);
          if( m_table[hi->m_hv].next(hi->m_iter,pair) ) {
            rv = TRUE;
            key = pair;
            break;
          }
        } while( hi->m_hv < m_table_size );
      }
      else {
        key = pair;
        rv = TRUE;
      }
    }

    return( rv );
  };

  virtual int prev( c_collection_iter<c_map_pair<K,V> > *iter, c_map_pair<K,V> &key ) const {
    hash_iter<K,V> *hi = (hash_iter<K,V>*)iter->m_data;
    int rv = FALSE;
    c_map_pair<K,V> pair;

    if( NULL == hi ) {
      rv = last(iter,key);
    }
    else {
      pair = key;
      if( !m_table[hi->m_hv].prev(hi->m_iter,pair) ) {
        do {
          // _DEBUG_COND( cout << "decrementing hash: " << hi->m_hv << endl ); 
          m_table[hi->m_hv--].complete_iter(hi->m_iter);
          if( hi->m_hv >= MAX_HV ) {
            rv = FALSE;
            break;
          }
          hi->m_iter - m_table[hi->m_hv].prepare_iter(iter->m_ord);
          if( m_table[hi->m_hv].next(hi->m_iter,pair) ) {
            rv = TRUE;
            key = pair;
            break;
          }
        } while( hi->m_hv >= 0 );
      }
    }
    return( rv );
  };

  virtual int last( c_collection_iter<c_map_pair<K,V> > *iter, c_map_pair<K,V> &key ) const {
    hash_iter<K,V> *hi = (hash_iter<K,V>*)iter->m_data;
    int rv = FALSE;
    c_map_pair<K,V> pair;
    if( iter->m_ord = pre_order ) {
      rv = first(iter,key);
    }
    else {
      if ( NULL == hi ) {
        hi = new hash_iter<K,V>;
        iter->m_data = hi;
        hi->m_hv = m_table_size -1;
        hi->m_iter = (c_collection_iter<c_map_pair<K,V> >*)NULL;
      }
      else {
        hi = (hash_iter<K,V>*)iter->m_data;
        if( NULL != hi->m_iter ) {
          m_table[hi->m_hv].complete_iter(hi->m_iter);
          // hi->m_iter = NULL;
          hi->m_iter = (c_collection_iter<c_map_pair<K,V> >*)NULL;
        }
      }

      hi->m_hv = m_table_size - 1;
      while( hi->m_hv != MAX_HV ) {
        hi->m_iter = m_table[hi->m_hv].prepare_iter(iter->m_ord);
        if( m_table[hi->m_hv].last(hi->m_iter,pair) ) {
          rv = TRUE;
          key = pair;
          break;
        }
        m_table[hi->m_hv].complete_iter(hi->m_iter );
        hi->m_hv--;
      }
    }
    return( rv );
  };

  virtual const c_hash<K,V>& operator=( const c_hash<K,V>& right ) {
    m_table_size = right.m_table_size;
    delete [] m_table;
    m_table = new c_map<K,V>[m_table_size];
    for( hash_value hv = 0; hv < m_table_size; hv++ ) {
      m_table[hv] = right.m_table[hv];
    }

    return *this;
  };

  virtual int operator==( const c_hash<K,V>& right ) {
    int rv = TRUE;

    if( get_count() != right.get_count() ) {
      return FALSE;
    }

    c_collection_iter<c_map_pair<K,V> > *iter = right.prepare_iter();
    c_map_pair<K,V> pair, my_pair;
    while( right.next( iter, pair ) ) {
      my_pair.m_key = pair.m_key;
      if( !find( my_pair ) ) {
        rv = FALSE;
        break;
      }

      if( !(my_pair.m_value == pair.m_value) ) {
        rv = FALSE;
        break;
      }
    }
    right.complete_iter(iter);

    return rv;
  };

  // new code added on: Tue Jun  2 16:51:23 EDT 1998 KRB
  virtual int remove( const K& key ) {
    hash_value hv = hash( m_table_size, key );
    return(m_table[hv].remove(key));
  };


private:
  c_map<K,V>    *m_table;
  hash_value    m_table_size;
};


// this algorithm was lifted from Algorithms in C++ Auth: Robert Sedgewick p233
inline hash_value hash( int m, const char* s )
{
  hash_value h = 0;
  for( int i = 0; s[i]; i++ ) 
    h = (64*h + s[i]) %m;
  return( h );
}

#endif // !defined( _C_HASH_H )



