#ifdef IGNORE_ME
  #define ident "@(#) c_list.h (/sccs/ri_util/include/s.c_list.h) 1.3 02/24/99 14:20:44";
#endif

/*  c_list.h
  Copyright 1997 Kyle R. Burton

This file is part of c_list.h, an attempt at a win32 equivalent for
the system log daemon available under unix.

c_list.h is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

c_list.h is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with c_list.h; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

=head1 NAME c_list - linked list object for C++

=head1 SYNOPSIS

  #include <c_list.h>
  ...
  c_list<int> primes;
  primes.add(4);
  primes.add(6);
  primes.add(8);
  primes.add(9);
  primes.add(10);

  cout << "there are: " << primes.get_count() << " prime nubers." << endl;

=head1 DESCRIPTION

This is a simple linked list object for C++.  It provides for it's own memory
management.

=head1 API REFERENCE

=head2 c_list()

Default constructor.  Takes no arguments.

=head2 c_list( const c_list<T>& that )

Copy constructor.  Creates a copy of the specifed list.

=head2 virtual ~c_list()

Destructor.  Frees all memory allocted by the list object.

=head2 virtual const c_list<T>& operator+=( const T& elm )

This member operates like an append operator, appending all
the elements from the second list onto the first.

=head2 virtual const c_list<T>& operator+=( const c_collection<T>& right )

This is a generic collection append operator.  

=head2 virtual const int operator==( const c_list<T>& right )

Equality operator.  Returns true if operator== succeeds for
all the elements in both lists.  Otherwise returns false.

=head2 virtual const c_list& operator=( const c_list<T>& right )

Assingment operator.  Copies a list object.

=head2 virtual int get_at( int pos, T& elm ) const

Retreives an element at a given position.  Sets elm equal
to the element at the specified position via T::operator==(const T&).
Returns true if successful, otherwise returns false.

=head2 virtual int set_at( int pos, const T& elm ) 

Sets an element at a given position.  Sets the element at the
specified position equal to elm via T::operator==(const T&).
Returns true if successful, otherwise returns false.

=head2 virtual T& operator[]( int pos )

Index operator.  Returns a reference to the element at the
given position.  If pos is invalid for the list object, it
throws a c_list_exception.

=head2 virtual void add_head( const T& elm )

Inserts a new element at the beginning of the list.

=head2 virtual void add_tail( const T& elm )

Appends a new element at the end of the list.

=head2 virtual int insert( int before, const T& elm )

Inserts a new element just before the index specified.

=head2 virtual int remove_head( T& elm )

Removes the first element from the list, and sets elm equal
to it.

=head2 virtual int remove_tail( T& elm )

Removes the last element from the list, and sets elm equal
to it.

=head2 virtual int remove( int num, T& elm )

Removes the element at the specified position, and sets
elm equal to it.

=head2 virtual int find( const T& fnd ) const

Attempts to find the first element in the list, starting
at the head of the list, for which T::operator== succeeds,
and reutrns the logical index of that element.  Returns -1
if it cannot find an element for which T::operator== succeeds.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

L<c_collection>

=cut

*/

/////////////////////////////////////////////////////////////////////////////////////////
// file:  c_list.h
//
// description: declaration, and implementation of template based c_list object
//
// creatd:  Sat 10/25/1997 14:34:17.26 Kyle R. Burton
//
#if !defined( _C_LIST )
#define _C_LIST 

#include <stdio.h>
#include "c_exception.h"
#include "c_collection.h"

class c_list_exception 
  : public c_exception
{
private:
  c_list_exception( const char* msg, char* file, int line ) 
    : c_exception( msg, file, line ) {
  };
public:
  ~c_list_exception() {
  };

  static c_list_exception* New( const char* msg, char* file, int line ) {
    return( new c_list_exception( msg, file, line ) );
  };

  virtual const c_list_exception& operator=( const c_list_exception& right ) {
    c_exception::operator=(right);
    return *this;
  };

  virtual const c_exception& operator=( const c_exception& right ) {
    c_exception::operator=(right);
    return *this;
  };
};

template<class T> class c_list;

typedef struct _list_node {
  void*      m_data;
  struct _list_node*  m_prev;
  struct _list_node*  m_next;
} list_node;


//
// class:  c_list
//
// description:  template based c_list class
//
template<class T>
class c_list 
  : virtual public c_collection<T>
{
public:
  //
  // default constructor
  //
  c_list() {
    init();
  };

  //
  // copy constructor
  //
  c_list( const c_list<T>& that ) {
    init();
    *this = that;
  };

  //
  // destructor
  //
  virtual ~c_list() {
    list_node *p = m_head;
    while( NULL != p ) {
      delete (T*)p->m_data;
      p = p->m_next;
      delete p->m_prev;
    }
  };


  virtual const c_list<T>& operator+=( const T& elm ) {
    add( elm );
    return(*this);
  };

  virtual const c_list<T>& operator+=( const c_collection<T>& right ) {
    c_collection_iter<T> *iter = right.prepare_iter( pre_order );
    T elm;

    while( right.next( iter, elm ) ) {
      add( elm );
    }
    right.complete_iter( iter );

    return(*this);
  };

  // implementation of the c_collection interface
  virtual int add( const T& elm ) {
    add_tail( elm );
    return( TRUE );
  };

  virtual int remove_all( void ) {
    T tmp;
    while( remove_head( tmp ) ) {
      ;// do nothing
    }
    return( TRUE );
  };

  virtual c_collection_iter<T>* prepare_iter( const collection_sorting_order ord = raw_order ) const {
    // return( new c_collection_iter<T>( c_collection<T>::owner_id(this), NULL, ord ) );
    return( new c_collection_iter<T>( owner_id(this), NULL, ord ) );
  };

  virtual int first( c_collection_iter<T>* iter, T& elm ) const { 
    // if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
    if( owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    int rv = FALSE;
    list_node *n;

    switch( iter->m_ord ) {
      case raw_order:
      case pre_order:
        n = m_head;
        if( n != NULL ) {
          elm = *((T*)n->m_data);
          iter->m_data = n;
          rv = TRUE;
        }
        else {
          iter->m_data = NULL;
        }
        break;
      case post_order:
        n = m_tail;
        if( n != NULL ) {
          elm = *((T*)n->m_data);
          iter->m_data = n;
          rv = TRUE;
        }
        else {
          iter->m_data = NULL;
        }
        break;
      default:
        throw( c_list_exception::New( "Invalid sort order", __FILE__, __LINE__ ) );
        // break;
    }

    return( rv );
  };
  
  virtual int next( c_collection_iter<T>* iter, T& elm ) const {
    // if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
    if( owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    int rv = FALSE;
    list_node *n = (list_node*)iter->m_data;;
    
    switch( iter->m_ord ) {
    case raw_order:
    case pre_order:
      if( NULL == n ) {
        return( first( iter, elm ) );
      }
      
      if( NULL != n->m_next ) {
        n = n->m_next;
        iter->m_data = n;
        elm = *((T*)n->m_data);
        rv = TRUE;
      }
      break;
    case post_order:
      if( NULL == n ) {
        return( last( iter, elm ) );
      }

      if( NULL != n->m_prev ) {
        n = n->m_prev;
        iter->m_data = n;
        elm = *((T*)n->m_data);
        rv = TRUE;
      }
      break;
    default:
      throw( c_list_exception::New( "Invalid sort order", __FILE__, __LINE__ ) );
      // break;
    }

    return( rv );
  };

  virtual int prev( c_collection_iter<T> *iter, T& elm ) const {
    // if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
    if( owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    int rv = FALSE;
    list_node *n = (list_node*)iter->m_data;
    
    switch( iter->m_ord ) {
    case post_order:
      if( NULL == n ) {
        return( first( iter, elm ) );
      }

      if( NULL != n->m_next ) {
        n = n->m_next;
        iter->m_data = n;
        elm = *((T*)n->m_data);
        rv = TRUE;
      }
      break;
    case raw_order:
    case pre_order:
      if( NULL == n ) {
        return( last( iter, elm ) );
      }

      if( NULL != n->m_prev ) {
        n = n->m_prev;
        iter->m_data = n;
        elm = *((T*)n->m_data);
        rv = TRUE;
      }
      break;
    default:
      throw( c_list_exception::New( "Invalid sort order", __FILE__, __LINE__ ) );
      // break;
    }

    return( rv );
  };

  virtual int last( c_collection_iter<T>* iter, T& elm ) const { 
    // if( c_collection<T>::owner_id(this) != iter->m_owner_id ) {
    if( owner_id(this) != iter->m_owner_id ) {
      return( FALSE );
    }

    int rv = FALSE;
    list_node *n;

    switch( iter->m_ord ) {
      case raw_order:
      case pre_order:
        n = m_head;
        if( n != NULL ) {
          elm = *((T*)n->m_data);
          iter->m_data = n;
          rv = TRUE;
        }
        else {
          iter->m_data = NULL;
        }
        break;
      case post_order:
        n = m_tail;
        if( n != NULL ) {
          elm = *((T*)n->m_data);
          iter->m_data = n;
          rv = TRUE;
        }
        break;
      default:
        throw( c_list_exception::New( "Invalid sort order", __FILE__, __LINE__ ) );
        // break;
    }

    return( rv );
  };

  virtual int complete_iter( c_collection_iter<T> *iter ) const {
    delete iter;
    return( TRUE );
  };

  virtual const int operator==( const c_list<T>& right ) {
    if( m_elms != right.m_elms ) {
      return( FALSE );
    }

    c_collection_iter<T> *iter = prepare_iter( pre_order );
    T elm;
    while( next( iter, elm ) ) {
      if( -1 == right.find( elm ) ) {
        return( FALSE );
      }
    }
    return( TRUE );
  };

  //
  // assingment operator
  //
  virtual const c_list& operator=( const c_list<T>& right ) {
    T elm;
    remove_all();

    c_collection_iter<T> *iter = right.prepare_iter();
    while( right.next( iter, elm ) ) {
      add( elm );
    }
    right.complete_iter( iter );

    return( *this );
  };

  //
  // sets elm equal to the element at pos, returns TRUE on success,
  // FALSE on failure
  //
  virtual int get_at( int pos, T& elm ) const {
    int rv = FALSE;
    list_node *tmp = _find( pos );

    if( NULL != tmp ) {
      elm = *((T*)tmp->m_data);
      rv = TRUE;
    }

    return( rv );
  };

  //
  // sets the elemnt at pos equal to elm, returns TRUE on success, FALSE
  // on failure
  //
  virtual int set_at( int pos, const T& elm ) {
    int rv = FALSE;
    list_node *tmp = _find( pos );

    if( NULL != tmp ) {
      *((T*)tmp->m_data) = elm;
      rv = TRUE;
    }

    return( rv );
  };

  //
  // returns refrence to the element at pos, throws exception of pos is invalid
  // for the list
  //
  virtual T& operator[]( int pos ) {
    if( pos < 0 || pos >= m_elms ) {
      char err_msg[256];
      sprintf( err_msg, "Invalid array index: %d", pos );
      throw( c_list_exception::New( err_msg, __FILE__, __LINE__ ) );
    }

    list_node *n = _find( pos );

    return( *((T*)n->m_data) );
  };

  //
  // adds a copy of elm as the head of the list
  //
  virtual void add_head( const T& elm ) {
    list_node *n = new list_node;
    n->m_data = (void*)new T(elm);
    n->m_prev = (list_node*)NULL;
    n->m_next = m_head;
    m_head = n;

    if( (list_node*)NULL == m_tail ) {
      m_tail = m_head;
    }
    else {
      n->m_next->m_prev = n;
    }

    m_elms++;
  };

  //
  // adds a copy of elm as the tail of the list
  //
  virtual void add_tail( const T& elm ) {
    list_node *n = new list_node;
    n->m_data = (void*)new T(elm);
    n->m_prev = m_tail;
    n->m_next = (list_node*)NULL;
    m_tail = n;

    if( (list_node*)NULL == m_head ) {
      m_head = m_tail;
    }
    else {
      n->m_prev->m_next = n;
    }

    m_elms++;
  };

  //
  // inserts a copy of elm into the list before the element at before
  // returns TRUE on success, FALSE on failure
  //
  virtual int insert( int before, const T& elm ) {
    if( before < 0 ) {
      add_head( elm );
      return( FALSE );
    }

    list_node *tmp = _find( before );

    if( NULL == tmp ) {
      return( FALSE );
    }

    list_node *n = new list_node;
    n->m_data = (void*)new T(elm);
    // attach to previous list_node
    n->m_prev = tmp->m_prev;
    n->m_prev->m_next = n;
    // attach to next list_node...
    n->m_next = tmp;
    tmp->m_prev = n;

    m_elms++;

    return( TRUE );
  };

  //
  // sets elm equal to the element at the head of the list, and removes the head element
  // from the list.  returns TRUE on success, FALSE on failure
  //
  virtual int remove_head( T& elm ) {
    list_node *n = m_head;
    int rv = FALSE;

    // no elms
    if( NULL != n ) {
      if( m_head->m_next != NULL ) {
        // more than 1 elm
        m_head = n->m_next;
        m_head->m_prev = (list_node*)NULL;
      }
      else {
        // this was the only elm in the list...
        m_head = (list_node*)NULL;
        m_tail = (list_node*)NULL;
        }
  
      elm = *((T*)n->m_data);
      delete (T*)n->m_data;
      delete n;
      m_elms--;
      rv = TRUE;
    }  
      return( rv );
  };

  //
  // sets elm equal to the tail element, removes the tail.  returns TRUE
  // on success, FALSE on failure.
  //
  virtual int remove_tail( T& elm ) {
    list_node *n = m_tail;
    int rv = FALSE;

    if( NULL != n ) {
      if( n->m_prev != NULL ) {
        // more than 1 elm
        m_tail = m_tail->m_prev;
        m_tail->m_next = (list_node*)NULL;
      }
      else {
        // only 1 elm...
        m_head = (list_node*)NULL;
      }

      elm = *((T*)n->m_data);
      delete (T*)n->m_data;
      delete n;
      m_elms--;

      rv = TRUE;
    }

    return( rv );
  };

  // 
  // finds the elemnt at num, sets elm equal to it, and removes it from the list.
  // returns TRUE on success, FALSE on failure
  //
  virtual int remove( int num, T& elm ) {
    list_node *n = _find( num );

    if( NULL == n ) {
      return( FALSE );
    }

    if( m_head == n ) {
      return( remove_head( elm ) );
    }

    if( m_tail == n ) {
      return( remove_tail( elm ) );
    }

    // attach prev to next
    n->m_prev->m_next = n->m_next;
    n->m_next->m_prev = n->m_prev;

    elm = *((T*)n->m_data);
    delete (T*)n->m_data;
    delete n;
    m_elms--;

    return( TRUE );
  };

  //
  // returns the number of elements in the list
  //
  virtual int get_count( void ) const {
    return( m_elms );
  };

  //
  // perofrms a linear search for fnd, and returns the index of the elemnt
  // that matches it, or -1 if not found
  //
  virtual int find( const T& fnd ) const {
    int pos = -1;
    list_node *n;

    for( int i = 0; i < m_elms; i++ ) {
      n = _find(i);
      if( fnd == *(T*)n->m_data ) {
        pos = i;
        break;
      }
    }

    return( pos );
  };

private:
  list_node*  m_head;
  list_node*  m_tail;
  list_node*  m_curr;
  int    m_elms;

  void init( void ) {
    m_head = (list_node*)NULL;
    m_tail = (list_node*)NULL;
    m_curr = (list_node*)NULL;
    m_elms = 0;
  };

  // accept a zero based pos
  list_node* _find( int pos ) const {
    if( pos < 0 || (pos+1) > m_elms ) {
      return( (list_node*)NULL );
    }

    list_node *tmp = m_head;
    while( pos-- ) {
      tmp = tmp->m_next;
    }

    return( tmp );
  };

};


#endif // !defined( _C_LIST )
