#ifdef IGNORE_ME
  #define ident "@(#) c_log_stream.h (/sccs/ri_util/include/s.c_log_stream.h) 1.3 02/24/99 14:20:45";
#endif

/*    c_log_stream.h
 *    Copyright 1998 Kyle R. Burton
 *
 * c_log_stream is free software; you can redistribute it and/or modify
 * it under the tearms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * c_log_stream is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with c_log_stream; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */      

/*

=head1 NAME

c_log_stream - utility to aid in redirecting output.

=head1 SYNOPSIS

  #incldue <c_log_stream.h>
  ...
  c_log_stream foo;

  foo.use_stdout();
  foo << "this goes to stdout." << endl;
  foo.use_file("/var/tmp/passwd");
  foo << "this goes to the file." << endl;

=head1 DESCRIPTION

This object if used in place of cout for logging, can make for an
easy transition to redirect output to an altenate stream.  Writing
your code to a c_log_stream object in leiu of cout directlry would
allow an easy one line (or cmdline switch) change to call use_file(),
and instantly have the remainder of the output from the program
redirected to the specified file.

=head1 API REFERENCE

=head2 c_log_stream() 

Default constructor.  Takes no arguments.

=head2 virtual ~c_log_stream() 

Destructor.  Closes any associated open filehandles.

=head2 virtual void use_stdout( void )

Redirects the log stream object to the tradiaional stdout,
unless it's already directed there.

=head2 virtual int use_file( const char* file, int shar = ios::app | ios::out  )

Attempts to open the given file name for writing.
All further data streamed to the log stream object will be
streamed into the given file.  Returns true on success, otherwise
returns false.

=head2 c_log_stream& operator<<( short s )

=head2 c_log_stream& operator<<( unsigned short s )

=head2 c_log_stream& operator<<( int s )

=head2 c_log_stream& operator<<( unsigned int s )

=head2 c_log_stream& operator<<( long s )

=head2 c_log_stream& operator<<( unsigned long s )

=head2 c_log_stream& operator<<( float s )

=head2 c_log_stream& operator<<( double s )

=head2 c_log_stream& operator<<( char s )

=head2 c_log_stream& operator<<( const char* s )

=head2 c_log_stream& operator<<( ostream&(s)(ostream&) )

These opreators are used to stream data into the log stream object.

=head2 ofstream& operator*(void)

Returns a reference to the internal ofstream object.  If you change
the state of the returned ofstream object, the further behavior
of stream object is undetermined.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

=cut

*/

#if !defined( _C_LOG_STREAM_H )
#define _C_LOG_STREAM_H

// this is just makes the operator<<() member functions easier to
// read, since they all do the same thing...
#define SHIFT_BODY    switch( disposition ) {\
                        case LOG:\
                          log << s;\
                          break;\
                        case STDOUT:\
                          cout << s;\
                          break;\
                        case STDERR:\
                          cerr << s;\
                          break;\
                      }

// act like cout, and ofstream...depending on our disposition
//   "Just who exactly do you want me to be?"
//   "Why, someone else of course."
class c_log_stream
{
public:
  enum output_location { LOG = 0, STDOUT, STDERR };

  // i considered being able to pass a disposition/filename to
  // the constructor, but then that created the need
  // to throw an exception...didn't think it was worth
  // all the bother...just call use_file() instead, and grok the
  // return value.
  c_log_stream() 
    : out(cout) 
    , err(cerr) 
    , disposition(STDOUT) 
    , file_is_open(FALSE) {
    ; // do nothing
  };

  // die.
  virtual ~c_log_stream() {
    if( file_is_open ) {
       log.close();
    }
  };

  // redirect to stdout from this point on...note that
  // this should close/flush any open log file.
  virtual void use_stdout( void ) {
    disposition = STDOUT;
    if( file_is_open ) {
      log.close();
      file_is_open = FALSE;
    }
  };

  // open the file for output (append is the default), and
  // set the disposition to cause output to stream to that file...
  virtual int use_file( const char* file, int shar = ios::app | ios::out  ) {
    if( file_is_open ) {
      file_is_open = FALSE;
      log.close();
    }

    log.open( file, shar );

    if( log.fail() ) {
      return FALSE;
    }

    disposition = LOG;
    file_is_open = TRUE;

    return TRUE;
  };

  // these functions should pretty much cover the gamut of basic
  // data types, so if you can << on some basic datatype to cout, you 
  // should pretty much be able to do it with c_log_stream...but I'm
  // sure more will have to be added later...
  c_log_stream& operator<<( short s )          { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( unsigned short s ) { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( int s )            { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( unsigned int s )   { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( long s )           { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( unsigned long s )  { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( float s )          { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( double s )         { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( char s )           { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( const char* s )    { SHIFT_BODY; return *this; };
  c_log_stream& operator<<( ostream&(s)(ostream&) ) { SHIFT_BODY; return *this; };

  // allows easy access to the log stream...
  ofstream& operator*(void) { return log; };

private:
  ostream&        out;
  ostream&        err;
  ofstream        log;
  output_location disposition;
  int             file_is_open;
};

#endif // !defined( _C_LOG_STREAM_H )
