#ifdef IGNORE_ME
  #define ident "@(#) c_map.h (/sccs/ri_util/include/s.c_map.h) 1.3 02/24/99 14:20:44";
#endif

  /*  c_map.h
  Copyright 1997 Kyle R. Burton

This file is part of c_map.h, an attempt at a win32 equivalent for
the system log daemon available under unix.

c_map.h is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

c_map.h is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with c_map.h; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

=head1 NAME

c_map - binary tree object.

=head1 SYNOPSIS

  #include <c_map.h>
  ...
  c_map<int,int> squares;
  squares.add(1,2);
  squares.add(2,5);
  squares.add(3,10);
  squares.add(4,17);

  int sq;
  squares.find(3,sq);
  cout << "The square of 3 is: " << sq << endl;

=head1 DESCRIPTION

c_map is a binary tree object suitable for storing 
arbitrary key value pairs, where uniqueness of keys 
is enforced.  The object manages it's own memory.
Stored objects must support the following members:

  a default constructor        T::T()
  a copy constructor           T::T(const T&)
  an assignment operator       T::operator=(const T&)
  an equality operator         T::operator==(const T&)

If they do not, any code that uses them within a c_map
will likely fail at compile time.

=head1 API REFERENCE

=head2 c_map() 

This is the default constructor.

=head2 c_map( const c_map<K,V>& that ) 

This is a copy constructor.  It creates a copy of
the given map object.

=head2 virtual ~c_map( )

Destructor, cleans up all memory allocated by the
c_map object.

=head2 virtual int clear_values( void )

When invoked, this member function will cause the c_map
object to walk it's internal binary tree, and invoke V::operator=()
on the values to a default V() object.

=head2 virtual int least( c_map_pair<KEY,VALUE>& elm ) const

This function sets the key and value members of the map pair
object to the key and value of the 'least', or leftmost element
in the tree.  It returns true on success, otherwise it returns
false.

=head2 virtual int greatest( c_map_pair<KEY,VALUE>& elm ) const

Retreives the 'greatest', or rightmost element in the tree.
It returns true on success, otherwise it returns false.

=head2 virtual int find( KEY &key ) const

This version of find() can be used to test for the existance
of a given key in the tree.
It returns true on success, otherwise it returns false.

=head2 virtual int find( c_map_pair<KEY,VALUE>& elm ) const

To retrive a value from the tree, first populate the key element
of the map pair, and then invoke this member function.  If the
key value pair is present in the tree, it will set the map pair's
value member equal to the found value.

=head2 virtual int is_key( const KEY& key ) const

Just as with the first version of find(), this function can 
be used to test for the existance
of a given key in the tree.
It returns true on success, otherwise it returns false.

=head2 virtual const c_map_pair<KEY,VALUE> key_find( const KEY& key ) const

This function attempts to locate a key/value pair in the tree, and
return a constant map pair that is initialized with the data at
the located node.
It returns true on success, otherwise it returns false.

=head2 virtual const c_map_pair<KEY,VALUE> val_find( const VALUE& val ) const

value_find() is very similar to key_find() in that it returns a
map pair filled with the key/value pair from the first node that contains
a match for val (via V::operator==).  Since the value is not what
the tree was ordered on, this function has to iterate through the tree,
node by node untill it finds a matching value.
It returns true on success, otherwise it returns false.

=head2 virtual int remove( const c_map_pair<KEY,VALUE>& val ) 

This function can be used to remove an element from the binary
tree.  You must first populate the key member of the map pair
to the key you wish removed from the tree before invoking
this member function.
It returns true on success, otherwise it returns false.

=head2 virtual int remove( const KEY& key )

This function attempts to remove the specified key from 
the map.
It returns true on success, otherwise it returns false.

=head2 virtual int operator==( const c_map<KEY,VALUE>& map ) const

This is an equality operator for c_map.  It compares each
element of the given trees.  It returns true if all elements
from both trees match, otherwise it returns false.

=head2 virtual insert_status insert( const KEY& key, const VALUE& value )

This member can be used to insert a key value pair into the tree.
It returns true on success, otherwise it returns false.

=head2 virtual int depth( void ) const

depth() was primarily added for diagnostic purposes, but can also
be used to help determine balance in the tree.  It traverses the
tree, and returns the greatest depth for a given branch of the tree.

=head2 virtual const c_map<KEY,VALUE>& operator=( const c_map<KEY,VALUE>& right )

This is the assignment operator.  It makes the tree on the
left side of the assignment a copy of the one on the right.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

L<c_collection>

Books on data structures, binary trees in specific.
Try Knuth, or Sedgwick, that's what was used as a 
basis for this code.

=cut

*/
//////////////////////////////////////////////////////////////////////////////////
// file: c_map.h
//
// description: declaration of the map class
//
// created: 10/27/97 Kyle R. Burton
//
// revision list:
//  12/24/97 KRB - fixed remove_tree() so it sets m_top to null after the delete...
//
#if !defined( _C_MAP_H )
#define _C_MAP_H

#include <iostream.h>

#include "c_collection.h"

class c_map_exception
  : virtual public c_exception
{
private:
  c_map_exception( const char* msg, char* file, int line ) 
    : c_exception( msg, file, line ) {
    ; // do nothing
  };
public:
  virtual ~c_map_exception() { ;/* do nothing */ };

  static c_map_exception* New( const char* msg, char* file, int line ) {
    return( new c_map_exception( msg, file, line ) );
  };

  virtual const c_exception& operator=( const c_exception& right ) {
    c_exception::operator=( right );
    return *this;
  };

  virtual const c_map_exception& operator=( const c_map_exception& right ) {
    c_exception::operator=( right );
    return *this;
  };
};

//////////////////////////////////////////////////////////////////////////////////
// class c_map
//
// c_map's purpose is to provide an efficient way of storing unique keys 
// and associated values
//

//
// class:  c_map_pair - used when setting/retreiving elements from
//        c_map...we need this so we can properly implement
//        the c_collection interface - where functions like
//        add() take a single argument...
//
template<class KEY, class VALUE>
class c_map_pair
{
public:
  c_map_pair() 
    : m_key()
    , m_value() {
  };

  c_map_pair( const KEY& key, const VALUE& val ) 
    : m_key( key )
    , m_value( val ) {
  };

  c_map_pair( const c_map_pair& that ) 
    : m_key( that.m_key )
    , m_value( that.m_value ) {
  };

  ~c_map_pair() {
    ;// do nothing
  };

  const KEY& key( void ) { return( m_key ); };
  const VALUE& value( void ) { return( m_value ); };

  virtual int operator==( const c_map_pair<KEY,VALUE>& right ) const {
    //return( (m_key == right.m_key) && (m_value == right.m_value) );
    return( (m_key == right.m_key) );
  };

  virtual int operator!=( const c_map_pair<KEY,VALUE>& right ) const {
    return( !(*this == right) );
  };

  virtual const c_map_pair<KEY,VALUE>& operator=( const c_map_pair<KEY,VALUE>& right ) {
    m_key = right.m_key;
    m_value = right.m_value;
    #if defined( __DEBUG )
      if( !(m_key == right.m_key && m_value == right.m_value) ) {
        cout << __FILE__ << "(" << __LINE__ << ") "
       << "Error c_map pair assingment went wrong." << endl;
        cout << __FILE__ << "(" << __LINE__ << ") "
       << m_key << " (keys) should equal: " << right.m_key << endl;
        cout << __FILE__ << "(" << __LINE__ << ") "
       << m_value << " (values) should equal: " << right.m_value << endl;
      }
    #endif
    return(*this);
  };

  KEY    m_key;
  VALUE  m_value;

private:
};


//
// class:  c_map
//
// only looks like a c_collection on the surface...
//
struct map_node {
  map_node(void)
    : m_key(0)
    , m_value(0)
    , m_left(0)
    , m_right(0)
    , m_parent(0) 
  {
    ;// do nothing...
  };

  void*      m_key;
  void*      m_value;
  struct map_node    *m_left;
  struct map_node    *m_right;
  struct map_node    *m_parent;
};

template<class KEY, class VALUE> class c_map 
  : virtual public c_collection<c_map_pair<KEY,VALUE> >
{
public:
  enum insert_status { status_error = 0, status_inserted, status_replaced };
  c_map( ) 
    : m_top(0)
    , m_elms(0) {
  };

  c_map( const c_map<KEY,VALUE>& that ) 
    : m_top(0)
    , m_elms(0) {
    *this = that;
  };

  virtual ~c_map( ) {
    remove_all();
  };

  // implement the c_collection interface...
  virtual int add( const KEY& elm ) {
    VALUE vtmp;
    insert_status rv = insert( elm, vtmp );
    return( rv != status_error );
  };

  virtual int add( const c_map_pair<KEY,VALUE > &elm ) {
    insert_status rv = insert(elm.m_key, elm.m_value);
    #if defined( __DEBUG )
      cout << __FILE__ << "(" << __LINE__ << ") "
           << "adding: (" << elm.m_key << "," << elm.m_value << ")" << endl;
    #endif
    return( rv != status_error );
  };

  virtual int remove_all( void ) {
    m_elms = 0;
    return( remove_tree( &m_top ) );
  };

  virtual int clear_values( void ) {
    return( _clear_values( &m_top ) );
  };

  virtual c_collection_iter<c_map_pair<KEY,VALUE> >* prepare_iter( const collection_sorting_order ord = raw_order ) const {
    // return( new c_collection_iter<c_map_pair<KEY,VALUE> >( c_collection<c_map_pair<KEY,VALUE> >::owner_id(this), NULL, ord ) );
    return( new c_collection_iter<c_map_pair<KEY,VALUE> >( owner_id(this), NULL, ord ) );
  };

  virtual int first( c_collection_iter<c_map_pair<KEY,VALUE> >* iter, c_map_pair<KEY,VALUE>& elm ) const { 
    int rv = FALSE;
    switch( iter->m_ord ) {
      case pre_order:
      case raw_order:
        iter->m_data = least( m_top );
        if( (map_node*)NULL != iter->m_data ) {
          elm.m_key = *((KEY*)((map_node*)iter->m_data)->m_key);
          elm.m_value = *((VALUE*)((map_node*)iter->m_data)->m_value);
          rv = TRUE;
        }
        break;
      case post_order:
        iter->m_data = greatest( m_top );
        if( (map_node*)NULL != iter->m_data ) {
          elm.m_key = *((KEY*)((map_node*)iter->m_data)->m_key);
          elm.m_value = *((VALUE*)((map_node*)iter->m_data)->m_value);
          rv = TRUE;
        }
        break;
      default:
        throw( c_map_exception::New( "invalid sort order", __FILE__, __LINE__ ) );
        // break;
    }

    return( rv );
  };

  virtual int next( c_collection_iter<c_map_pair<KEY,VALUE> >* iter, c_map_pair<KEY,VALUE>& elm ) const {
    map_node* tmp = (map_node*)NULL;
    int rv = FALSE;
    switch( iter->m_ord ) {
      case pre_order:
      case raw_order:
        tmp = next( (map_node*)iter->m_data );
        if( (map_node*)NULL != tmp ) {
          iter->m_data = tmp;
          elm.m_key  = *((KEY*)((map_node*)iter->m_data)->m_key);
          elm.m_value  = *((VALUE*)((map_node*)iter->m_data)->m_value);
          rv = TRUE;
        }
        else {
          if( NULL == iter->m_data ) {
            rv = first( iter, elm );
          }
        }
        break;
      case post_order:
        tmp = prev( (map_node*)iter->m_data );
        if( (map_node*)NULL != tmp ) {
          iter->m_data = tmp;
          elm.m_key  = *((KEY*)((map_node*)iter->m_data)->m_key);
          elm.m_value  = *((VALUE*)((map_node*)iter->m_data)->m_value);
          rv = TRUE;
        }
        else {
          if( NULL == iter->m_data ) {
            return( last( iter, elm ) );
          }
        }
        break;
      default:
        throw( c_map_exception::New( "invalid sort order", __FILE__, __LINE__ ) );
        // break;
    }

    return( rv );
  };

  virtual int prev( c_collection_iter<c_map_pair<KEY,VALUE> >* iter, c_map_pair<KEY,VALUE>& elm ) const { 
    map_node* tmp = (map_node*)NULL;
    int rv = FALSE;
    switch( iter->m_ord ) {
      case pre_order:
      case raw_order:
        tmp = prev( (map_node*)iter->m_data );
        if( (map_node*)NULL != tmp ) {
          iter->m_data = tmp;
          elm.m_key  = *((KEY*)((map_node*)iter->m_data)->m_key);
          elm.m_value  = *((VALUE*)((map_node*)iter->m_data)->m_value);
          rv = TRUE;
        }
        else {
          if( NULL == iter->m_data ) {
            rv = last( iter, elm );
          }
        }
        break;
      case post_order:
        tmp = next( (map_node*)iter->m_data );
        if( (map_node*)NULL != tmp ) {
          iter->m_data = tmp;
          elm.m_key  = *((KEY*)((map_node*)iter->m_data)->m_key);
          elm.m_value  = *((VALUE*)((map_node*)iter->m_data)->m_value);
          rv = TRUE;
        }
        else {
          if( NULL == iter->m_data ) {
            return( last( iter, elm ) );
          }
        }
        break;
      default:
        throw( c_map_exception::New( "invalid sort order", __FILE__, __LINE__ ) );
        // break;
    }

    return( rv );
  };
  
  virtual int last( c_collection_iter<c_map_pair<KEY,VALUE> >* iter, c_map_pair<KEY,VALUE>& elm ) const { 
    switch( iter->m_ord ) {
      case pre_order:
      case raw_order:
        iter->m_data = greatest( m_top );
        elm.m_key = *((KEY*)((map_node*)iter->m_data)->m_key);
        elm.m_value = *((VALUE*)((map_node*)iter->m_data)->m_value);
        break;
      case post_order:
        iter->m_data = least( m_top );
        elm.m_key = *((KEY*)((map_node*)iter->m_data)->m_key);
        elm.m_value = *((VALUE*)((map_node*)iter->m_data)->m_value);
        break;
      default:
        throw( c_map_exception::New( "invalid sort order", __FILE__, __LINE__ ) );
        // break;
    }

    return( TRUE );
  };

  virtual int complete_iter( c_collection_iter<c_map_pair<KEY,VALUE> >* iter ) const {
    delete iter;
    return( TRUE );
  };


  virtual int least( c_map_pair<KEY,VALUE>& elm ) const {
    map_node *tmp = least( m_top );
    int rv = FALSE;

    if( (map_node*)NULL != tmp ) {
      elm.m_key = *((KEY*)tmp->m_key);
      elm.m_value = *((VALUE*)tmp->m_value);
      rv = TRUE;
    }

    return( rv );
  };

  virtual int greatest( c_map_pair<KEY,VALUE>& elm ) const {
    map_node *tmp = greatest( m_top );
    int rv = FALSE;

    if( (map_node*)NULL != tmp ) {
      elm.m_key = *((KEY*)tmp->m_key);
      elm.m_value = *((VALUE*)tmp->m_value);
      rv = TRUE;
    }

    return( rv );
  };

  virtual int find( KEY &key ) const {
    map_node *tmp = kfind( key );
    if( (map_node*)NULL != tmp ) {
      key = *((KEY*)tmp->m_key);
    }
    return( NULL != tmp );
  };

  virtual int find( c_map_pair<KEY,VALUE>& elm ) const {
    map_node *tmp = kfind( elm.m_key );
    int rv = FALSE;

    if( (map_node*)NULL != tmp ) {
      elm.m_value = *((VALUE*)tmp->m_value);
      rv = TRUE;
    }

    return( rv );
  };

  virtual int is_key( const KEY& key ) const {
    return( NULL != kfind(key) );
  };

  virtual const c_map_pair<KEY,VALUE> key_find( const KEY& key ) const {
    map_node* tmp = kfind(key);
    c_map_pair<KEY,VALUE> pair;

    if( (map_node*)NULL != tmp ) {
      pair.m_key = *((KEY*)tmp->m_key);
      pair.m_value = *((VALUE*)tmp->m_value);
    }

    return( pair );
  };

  virtual const c_map_pair<KEY,VALUE> val_find( const VALUE& val ) const {
    map_node* tmp = vfind(val);
    c_map_pair<KEY,VALUE> pair;

    if( (map_node*)NULL != tmp ) {
      pair.m_key = *((KEY*)tmp->m_key);
      pair.m_value = *((VALUE*)tmp->m_value);
    }

    return( pair );
  };

  virtual int remove( const c_map_pair<KEY,VALUE>& val ) {
    return( remove( val.m_key ) );
  };

  // complete re-write of this function on: Thu Jun  4 08:23:05 EDT 1998 KRB
  // remove a key/value pair from the tree  this is _much_ more complex
  // than I had at first assumed...
  virtual int remove( const KEY& key ) {
    map_node* tmp = kfind(key);

    if( (map_node*)NULL == tmp ) {
      // no matching node found...
      return FALSE;
    }

    map_node *left   = tmp->m_left,
             *right  = tmp->m_right,
             *parent = tmp->m_parent;

    if( m_top == tmp ) {
      // it's the parent node, stick the left child-node into the
      // right child's tree
      // at this point there can be a few different situations:
      //  - there are no children
      //  - there are both a leaf and a right
      //  - there is 1 child, but not both (either right or left)

      // both children are null -- kill the tree
      if( (map_node*)NULL == right && (map_node*)NULL == left ) {
        m_top = (map_node*)NULL;
      }
      else if( (map_node*)NULL != right && (map_node*)NULL != left ) {
        // there are child trees both on the left, and on the right
        // prune the right's parent, and set the right child as the new top
        right->m_parent = (map_node*)NULL;
        m_top = right;

        // prune the left's parent, add the left to the right's tree
        left->m_parent = (map_node*)NULL;
        add_node( left, &right );
      }
      else {
        // there's only one child is it the left or right?
        map_node *p = (map_node*)NULL == right ? left : right;

        // prune it's parent, and make it the new top node...
        p->m_parent = (map_node*)NULL;
        m_top = p;
      }
    }
    else {
      // tmp is a node somewhere inside the tree, go through
      // the process of placing the children where we are...
      // There are a few situations here:
      // -- tmp has no children
      // -- tmp has both children
      // -- tmp has 1 child (left or right)

      if( (map_node*)NULL == right && (map_node*)NULL == left ) {
        // no children, just prune the parent's pointer to tmp
        if( parent->m_left == tmp ) {
          parent->m_left = (map_node*)NULL;
        }
        else if( parent->m_right == tmp ) {
          parent->m_right = (map_node*)NULL;
        }
        else {
          cout << __FILE__ << "(" << __LINE__ << ") Erorr in c_map::remove() - child node"
               << " was neither the left, or right node of it's parent? " 
               << " hmm...how'd you manage to get here?"  << endl;
        }
      }
      else if( (map_node*)NULL != right && (map_node*)NULL != left ) {
        // both children exist
        // place right where we are
        if( tmp == parent->m_left ) {
          parent->m_left = right;
          right->m_parent = parent;
        }
        else {
          parent->m_right = right;
          right->m_parent = parent;
        }

        // now we need to prune left's parent, and stick it into the tree...
        left->m_parent = (map_node*)NULL;
        add_node( left, &right );
      }
      else {
        // there is only 1 child, put it where we are
        map_node *child = (map_node*)NULL == left ? right : left;
        child->m_parent = parent;
        if( parent->m_left == tmp ) {
          parent->m_left = child;
        }
        else {
          parent->m_right = child;
        }
      }
    }

    // free the orphaned node, and it's data...
    delete (KEY*)tmp->m_key;
    delete (VALUE*)tmp->m_value;
    delete tmp;

    return TRUE;
  };

  virtual int operator==( const c_map<KEY,VALUE>& map ) const {
    int rv = TRUE;

    if( map.m_elms != m_elms ) {
      return( FALSE );
    }

    c_collection_iter<c_map_pair<KEY,VALUE> > *iter = prepare_iter( raw_order );
    c_map_pair<KEY,VALUE> our_pair, their_pair;
    while( next( iter, our_pair ) ) {
      their_pair = our_pair;
      if( !map.find( their_pair ) ) {
        rv = FALSE;
        break;
      }

      if( our_pair != their_pair ) {
        rv = FALSE;
        break;
      }
    }
    complete_iter( iter );

    return( rv );
  };

  virtual int get_count( void ) const {
    return( m_elms );
  };

  virtual insert_status insert( const KEY& key, const VALUE& value ) {
    map_node* tmp = new map_node;

    tmp->m_key = new KEY(key);
    tmp->m_value = new VALUE(value);

    return( add_node( tmp, &m_top ) );
  };

  virtual int depth( void ) const {
    return( _depth( m_top ) );
  };

  map_node* copy_tree( void ) const {
    return( copy_tree( m_top ) );
  };

  static int delete_tree( map_node** top ) {
    return( remove_tree( top ) );
  };

  virtual const c_map<KEY,VALUE>& operator=( const c_map<KEY,VALUE>& right ) {
    remove_all();
    m_top = copy_tree(right.m_top);
    m_elms = right.m_elms;
    return *this;
  };

  friend class c_str_map_iter;
private:
  map_node*     m_top;
  int      m_elms;

  // _HEY_YOU_: did you forget to implement
  //  const KEY& operator=(const&KEY)?
  static int direction( const KEY& left, const KEY& right ) {
    return( (left == right) ? 0 : ((left < right) ? -1 : 1 ) );
  };

  map_node* vfind( const VALUE& val ) const {
    map_node* tmp = least( m_top );
    VALUE* pval;

    while( (map_node*)NULL != tmp ) {
      pval = (VALUE*)tmp->m_value;
      if( *pval == val ) {
        break;
      }

      tmp = next( tmp );
    }

    return( tmp );
  };

  map_node* kfind( const KEY& key ) const {
    int val = 0;
    map_node* tmp = m_top;

    if( (map_node*)NULL == tmp ) {
      return( (map_node*)NULL );
    }

    while( (map_node*)NULL != tmp ) {
      KEY test = *((KEY*)tmp->m_key);
      val = direction( key, test );

      switch( val ) {
        case 0:    // equal to
          return( tmp );
          // break;
        case -1:  // key is less, go left...
          tmp = tmp->m_left;
          break;
        case 1:    // key is greater, go right...
          tmp = tmp->m_right;
          break;
        default:  // error
          throw( c_map_exception::New( "error during key compare", __FILE__, __LINE__ ) );
          // break;
      }
    }

    return( tmp );
  };

  insert_status add_node( map_node* n, map_node** ppos ) {
    insert_status rv = status_error;
    
    if( (map_node*)NULL == n || (map_node**)NULL == ppos ) {
      return( rv );
    }

    if( (map_node*)NULL == *ppos ) {
      *ppos = n;
      m_elms++;
      return( status_inserted );
    }

    map_node *pos = *ppos;
    while( (map_node*)NULL != pos ) {
      int val = direction( *((KEY*)n->m_key), *((KEY*)pos->m_key) );
      switch( val ) {
        case 0:    // here
          // in this case we're adding a node to a place where it already
          // exists -- assign the value, and delete the node...
          *((VALUE*)pos->m_value) = *((VALUE*)n->m_value);
          delete (KEY*)n->m_key;
          n->m_key = NULL;
          delete (VALUE*)n->m_value;
          n->m_value = NULL;
          delete n;
          pos  = (map_node*)NULL;
          rv = status_replaced;
          break;
        case 1:    // right
          if( NULL != pos->m_right ) {
            pos = pos->m_right;
          }
          else {
            n->m_parent = pos;
            pos->m_right = n;
            pos = (map_node*)NULL;
            m_elms++;
            rv = status_inserted;
          }
          break;
        case -1:  // left
          if( NULL != pos->m_left ) {
            pos = pos->m_left;
          }
          else {
            n->m_parent = pos;
            pos->m_left = n;
            pos = (map_node*)NULL;
            m_elms++;
            rv = status_inserted;
          }
          break;
        default:
          throw( c_map_exception::New( "error determining direction", __FILE__, __LINE__ ) );
          // break;
      }
    }

    return( rv );
  };

  static map_node* least( map_node* top ) {
    map_node *tmp = top;

    if( (map_node*)NULL == tmp ) {
      return( tmp );
    }

    while( (map_node*)NULL != tmp->m_left ) {
      tmp = tmp->m_left;
    }

    return( tmp );
  };

  static map_node* greatest( map_node* top ) {
    map_node *tmp = top;

    if( (map_node*)NULL == tmp ) {
      return( tmp );
    }

    while( (map_node*)NULL != tmp->m_right ) {
      tmp = tmp->m_right;
    }

    return( tmp );
  };

  static map_node* next( map_node* elm ) {
    map_node *tmp = elm;
    if( (map_node*)NULL == tmp ) {
      return( (map_node*)NULL );
    }

    if( NULL != tmp->m_right ) {
      // go right, and fall left...
      tmp = tmp->m_right;
      while( (map_node*)NULL != tmp->m_left ){
        tmp = tmp->m_left;
      }

      return( tmp );
    }

    if( (map_node*)NULL == tmp->m_parent ) {
      return( (map_node*)NULL );
    }

    if( tmp->m_parent->m_left != tmp ) {
      // find the next parent map_node that we're not a right child of...
      while( TRUE ) {
        if( NULL == tmp->m_parent ) {
          break;
        }

        if( tmp->m_parent->m_left == tmp ) {
          break;
        }

        tmp = tmp->m_parent;
      }
    }

    // we're our parent's left child...our parent is next...
    return( tmp->m_parent );
  };

  static map_node* prev( map_node* elm ) {
    if( (map_node*)NULL == elm ) {
      return( (map_node*)NULL );
    }

    if( NULL != elm->m_left ) {
      // go left, and fall right...
      elm = elm->m_left;
      while( (map_node*)NULL != elm->m_right ) {
        elm = elm->m_right;
      }

      return( elm );
    }

    if( (map_node*)NULL == elm->m_parent ) {
      return( (map_node*)NULL );
    }

    if( elm->m_parent->m_right != elm ) {
      while( TRUE ) {
        if( (map_node*)NULL == elm->m_parent ) {
          break;
        }

        if( elm->m_parent->m_right == elm ) {
          break;
        }

        elm = elm->m_right;
      }
    }

    return( elm->m_parent );
  };

  static int remove_tree( map_node** top ) {
    int rv = FALSE;

    if( (map_node*)NULL != *top ) {
      remove_tree( &(*top)->m_left );
      remove_tree( &(*top)->m_right );
  
      delete (KEY*)(*top)->m_key;
      (*top)->m_key = (KEY*)NULL;
      delete (VALUE*)(*top)->m_value;
      (*top)->m_value = (VALUE*)NULL;
      delete (*top);
      *top = (map_node*)NULL;
      rv = TRUE;
    }

    return( rv );
  };

  static int _clear_values( map_node** top ) {
    int rv = FALSE;

    if( (map_node*)NULL != *top ) {
      _clear_values( &(*top)->m_left );
      _clear_values( &(*top)->m_right );

      (*(VALUE*)(*top)->m_value) = VALUE();
      rv = TRUE;
    }

    return rv;
  };

  static int _depth( map_node* n = (map_node*)NULL ) {
    int d = 0;
    if( (map_node*)NULL != n ) {
      d = _depth( n->m_left );
      int tmp = _depth( n->m_right );
      d = (d > tmp) ? d : tmp;
      d++;
    }

    return( d );
  };

  static map_node* copy_tree( map_node* t ) {
    map_node *n = (map_node*)NULL;

    if( (map_node*)NULL != t ) {
      n = new map_node;
      n->m_key = new KEY(*((KEY*)t->m_key));
      n->m_value = new VALUE(*((VALUE*)t->m_value));

      n->m_left = copy_tree( t->m_left );
      n->m_right = copy_tree( t->m_right );
    }

    return( n );
  };

};

#endif // _C_MAP_H
