#ifdef IGNORE_ME
  #define ident "@(#) c_mutex.h (/sccs/ri_util/include/s.c_mutex.h) 1.6 02/24/99 14:20:44";
#endif


/* 
 * file:   c_mutex.h
 *
 * descirption:  implements mutex, and locking objects...
 *               an attempt will be made to make this platform independant,
 *               though the first target platform will be HP-UX
 *
 * reivsion list:
 *    Wed Jun 17 09:23:14 EDT 1998 KRB -- created the file
 *    Thu Jun 18 09:04:26 EDT 1998 KRB -- finally got it to work, at least
 *      it does on HP-UX...
 *
 */

/*

=head1 NAME

c_mutex - mutex object

=head1 SYNOPSIS

  #include <c_mutex.h>
  ...
  c_mutex protect_this_resource;
  ...
  void protect_me( void ) 
  {
    c_mutex_lock locked(protect_this_resource);
    // use some shared resource
  }

  void protect_me_too( void ) 
  {
    c_mutex_lock locked(protect_this_resource);
    // use some shared resource
  }

=head1 DESCRIPTION

The mutex, and mutex lock, objects provide simplified protection,
for resources that require mutual exclusion.

There are two ojbects used here, a c_mutex object, and a
c_mutex_lock object.  The mutex object is what get's shared
by the threads that need to protect the resource.  The mutex
lock object is used to obatin a lock on the mutex.  The
normal usage is to construct the lock object inside the scope
where the lock is necessary, and allow the lock object to destroy
itself (and thusly releasing the lock) when it goes out of scope.

This object uses the standard Unix semaphores in a singlular
manner, not the POSIX mutexes.  This menas that a file is
created in the file system that is associated with the mutex
object.  It will only be cleaned up if the mutex object gets
the chance to destruct itself.  Additionaly, if the object 
does not get the chance to clean up after itself (i.e. destruct),
it will not be able to free the allocated semaphore.

All the normal issues with multi threading and use of mutexes
applies.

=head1 API REFERENCE - c_mutex

=head1 c_mutex( ostream& o = cout ) 

Default constructor.  Takes an ostream object for redirection
of output.

=head2 virtual ~c_mutex()

Destructor.  Releases the allocated semaphore.

=head1 API REFERENCE - c_mutex_lock

=head2 c_mutex_lock( c_mutex& mtx, ostream& o = cout ) 

Constructor.  Takes a reference to the mutex object, to 
perform the lock upon, and performs the lock.  The construction
of this object will block untill the lock is obtained.

=head2 ~c_mutex_lock()

Destructor for the lock object.  Releases the lock
on the mutex object passed to the constructor.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

Books on multi-threaded programming concepts.  Try Pthreads 
Programming for starters.

IPC documentation for your system.

=cut

*/

#if !defined( _RC_MUTEX_H )
#define _RC_MUTEX_H

#include <iostream.h>
#include <stdio.h>

// these are appropriate for HP-UX, other platforms will need to
// be handeled as appropraite...
  #include <sys/types.h>
  #include <sys/ipc.h>
  #include <sys/sem.h>
  #include <unistd.h>

#include <stdinc.h>
#include <c_timestamp.h>
#include <c_exception.h>


#define MUTEX_FILE_BUFF_SIZE 256
#define MUTEX_OBJECT_NAME    "c_mutex"
#define MUTEX_FTOK_ID        'M'
// perhaps this should be a global temp directory?
#define MUTEX_PATH           "."

class c_mutex {
public:
  c_mutex( ostream& o = cout ) 
    : state(0) 
    , semid(-1) 
    , semkey(0)
    , out(o) 
    , fp(NULL) {

    // create a unique file name...
    pid_t pid = getpid();
    sprintf(file, "%s.%s.%08x.%08x", MUTEX_PATH, MUTEX_OBJECT_NAME, pid, this);
    #if defined( _DEBUG )
      INFO(out) << "calling ftok() with: " << file << endl << flush;
    #endif

    // now we need to open the file...
    fp = fopen( file, "w" );
    if( NULL == fp ) {
      ERROR(out) << "Erorr opening the file for the ftok() call..." << endl;
      throw( c_exception( 
         "Erorr opening the file for the ftok() call...",
         __FILE__,
         __LINE__ )
      );
    }

    // make the call to fok()...
    #if defined( _DEBUG )
      INFO(out) << "calling: ftok(" << file << "," << MUTEX_FTOK_ID << ")" << endl;
    #endif
    semkey = ftok(file,MUTEX_FTOK_ID);

    // allocate the semaphore...
    #if defined( _DEBUG )
      INFO(out) << "calling: semget( " << semkey << ", 1, 0600 )" << endl;;
    #endif
    semid = semget( semkey, 1, IPC_CREAT | 0600 );
    #if defined( _DEBUG )
      INFO(out) << "semget returned: " << semid << endl;
    #endif
    if( -1 == semid ) {
      fclose(fp);
      ERROR(out) << "Erorr allocating semaphore: semget() returned -1" << endl;
      throw( c_exception( "Erorr allocating semaphore: semget() returned -1",
             __FILE__, __LINE__ )
      );
    }
  };

  virtual ~c_mutex() {
    #if defined( _DEBUG )
      out << "~c_mutex()" << endl << flush;
    #endif

    #if defined( _DEBUG )
      out << "~c_mutex() closing file: " << file << endl << flush;
    #endif

    // release the semaphore...but only if we can get a lock on it...
    fclose(fp);

    #if defined( _DEBUG )
      out << "~c_mutex() closed file: " << file << endl << flush;
    #endif
    int rv = -1;

    #if defined( _DEBUG )
      out << "~c_mutex() calling semctl()" << endl << flush;
    #endif

    #if defined( _HP_UX )
      rv = semctl( semid, 1, IPC_RMID );
    #elif defined( _LINUX )
      union semun smn;
      rv = semctl( semid, 1, IPC_RMID, smn );
    #endif

    if( rv == -1 ) { // -1 for linux...
      ERROR(out) << "Error deleting semaphore." << endl;
      throw( c_exception( "Erorr deleting semaphore!", __FILE__, __LINE__ )
      );
    }

    // delete the file too...
    #if defined( _DEBUG )
      out << "~c_mutex() unlinking file: " << file << endl << flush;
    #endif
    unlink(file);
  };

friend class c_mutex_lock;

private:
  int       semid;
  key_t     semkey;
  int       state;
  ostream&  out;
  char      file[MUTEX_FILE_BUFF_SIZE];
  FILE*     fp;
};


// all locking is performed through this object...create it, passing a refrence to a
// mutex objec, and it get's locked, when this object get's destoryed, the
// lock is released...
class c_mutex_lock {
public:
  c_mutex_lock( c_mutex& mtx, ostream& o = cout ) 
    : mutex(mtx) 
    , out(o) {
    lock_buff[0].sem_num = 0;
    lock_buff[0].sem_op  = 0;
    lock_buff[0].sem_flg = 0;

    lock_buff[1].sem_num = 0;
    lock_buff[1].sem_op  = 1;
    lock_buff[1].sem_flg = 0;

    unlock_buff[0].sem_num = 0;
    unlock_buff[0].sem_op  = -1;
    unlock_buff[0].sem_flg = IPC_NOWAIT;

    // obtain the lock...
    if( semop( mutex.semid, &lock_buff[0], 2 ) < 0 ) {
      ERROR(out) << "Erorr allocating semaphore: semget() returned -1" << endl;
      throw( c_exception( "Erorr allocating semaphore: semget() returned -1",
              __FILE__, __LINE__ )
      );
    }
    #if defined( _DEBUG )
      INFO(out) << "[c_mutex_lock] obtained lock!" << endl;
    #endif
  };

  ~c_mutex_lock(void) {
    // release the lock...
    if( semop( mutex.semid, &unlock_buff[0], 1 ) < 0 ) {
      ERROR(out) << "Erorr allocating semaphore: semget() returned -1" << endl;
      throw( c_exception( "Erorr allocating semaphore: semget() returned -1",
        __FILE__, __LINE__ )
      );
    }
    #if defined( _DEBUG )
      INFO(out) << "[c_mutex_lock] lock released!" << endl;
    #endif
  };

private:
  c_mutex&      mutex;
  struct sembuf  lock_buff[2];
  struct sembuf  unlock_buff[1];
  ostream&       out;
};

#endif // defined( _RC_MUTEX_H )
