/*
 *
 * file:        c_perl_wrapper.h
 *
 * description: wrapper class to make the un-threadsafe perl 5.004 
 *    thread safe for our environments in C++.
 *  ALL use of perl should be through this object.
 *
 * reiviosn list:
 *    Tue Jan  5 12:49:59 EST 1999 KRB/CFAN - created the file
 *
 */

/*

=head1 NAME

c_perl_wrapper - C++ wrapper object for Perl 5.004

=head1 SYNOPSIS

  #include <c_perl_wrapper.h>

  char **argv = { "init_code.pm",NULL };
  c_string error;
  c_perl_wrapper pw;
  c_list<c_string> in_args, out_args;

  pw.setup_perl_wrapper(1,argv);
  in_args.add(c_string("arg1"));
  in_args.add(c_string("arg2"));
  pw.run_perl_sub(sub,in_args,out_args,error);
  pw.eval_string("print \"this is in perl\n\";");

=head1 DESCRIPTION

This object was created to protect threaded programs from the aspects
of the Perl 5.004 embedded libraries that are not thread safe.  It provides
a simplified interface to the embedded Perl functionality, as well as 
mutex locking to prevent threads from clobbering static data areas.

In addition to the benefits described above, using this object also keeps
the EXTERN.h, and perl.h headers out of your code.  We have experienced 
limitations with the HP-UX C++ compiler not being albe to properly compile
a module with a main() function, and problems with the pre-processor and
'header overload', where the compiler dies without warning.

=head1 API REFERENCE

=head2 c_perl_wrapper::c_perl_wrapper()

Default, and only constructor.  Takes no arguments.

=head2 virtual c_perl_wrapper::~c_perl_wrapper()

Destructor.  If the object has been initialized, it will clean up
any allocated perl interpreter.

=head2 virtual int c_perl_wrapper::setup_perl_wrapper(int,char**)

This function allocates the perl interpreter, and gives the caller 
the opportunity to pass command line arguments to the interpreter during
initilalization.  This can be used to load functions from a module, or 
run a script.

=head2 virtual perl_rv c_perl_wrapper::perl_rv run_perl_sub()

This function can be called after initialization, to invoke a perl
subroutine.  It takes three arguments:  
  - a c_string speicifying the name of the subroutine to run
  - a c_list<c_string> of the input arguments
  - a reference to a c_list<c_string> to receive the output arguments

=head2 virtual int c_perl_wrapper::eval_string(const char*)

eval_string() can be used to evaluate arbritrary perl code.  It takes
one arguments, a const char* to a string that holds the code to be 
evaluated.

Currently there is no mechanisim for passing or returning arguments to 
or from the arbritrary code.

=head2 virtual int c_perl_wrapper::get_perl_error(c_string&)

This can be used to retreive the last error that occurred.  The
wrapper object stores any errors that Perl reports in an internal
buffer, you can use get_perl_error to retreive them at will.

=head1 REVISION LIST

Wed Feb  3 14:48:57 EST 1999 KRB - added the POD documentation.

=head1 BUGS

No known bugs.

=head1 AUTHORS

  Cindy Fan cfan@rosenbluth.com 215.977.4684
  Kyle R. Burton kburton@rosenbluth.com 215.977.4630

=cut

 */

#ifdef IGNORE_ME
  #define ident "@(#) c_perl_wrapper.h (/sccs/ri_util/include/s.c_perl_wrapper.h) 1.6 02/24/99 14:20:45";
#endif

#include <c_string.h>
#include <c_list.h>
#include <c_mutex.h>
#if !defined( _RC_PERL_WRAPPER )
#define _RC_PERL_WRAPPER 

class c_perl_wrapper {
public:
  enum perl_rv { FAILURE = 0, UNDEF, SCALAR, ARRAY };
  enum wrapper_state { UNINITIALIZED = 0, ERROR, INITIALIZED };

  c_perl_wrapper();
  virtual ~c_perl_wrapper();
  virtual int setup_perl_wrapper(int argc, char** argv);

  virtual perl_rv run_perl_sub(
    const c_string& sub_name,
    c_list<c_string>& in_args, 
    c_list<c_string>& out_args
  );

  virtual int eval_string(char* perl_code);
	virtual int get_scalar(char* var,c_string& val);
  virtual int get_perl_error(c_string& e);

private:
  void                   *perl_interpreter; // this is needed by the perl libs
  static c_mutex         perl_mutex;        // for protecting the non-threadsafe calls
  c_string               perl_error;

	wrapper_state          state;
};

#endif
