#ifdef IGNORE_ME
  #define ident "@(#) c_socket.h (/sccs/ri_util/include/s.c_socket.h) 1.7 02/24/99 14:20:44";
#endif

/*
 * file: c_socket.h
 *
 * description:  this is an attempt at a platform independent 
 *    socket implementation
 *
 * revision list: 
 *   Tue Jun 23 09:49:39 EDT 1998 KRB - created
 *   Thu Sep 10 08:41:42 EDT 1998 KRB - added prototypes for getpeeraddr,
 *     and getsockaddr.  moved name_resolve into the puublic area
 *
 *
 *
 *
 */

/*

=head1 NAME

c_socket - tcp socket object

=head1 SYNOPSIS

  #include <c_socket.h>
  ...
  c_socket sck;
  sck.create_socket();
  sck.connect(ip_addr,port_num);
  int size = 10;
  sck.send("some data",size),

=head1 DESCRIPTION

tcp/ip socket object.  This object provides basic services,
and some additional functionality, such as line reading, for sockets.
It can be used as either the client side (initiator), or
server side (accepter) of a socket connection.

=head1 API REFERENCE

=head2 c_tcp_socket( ostream& o = cout )

Constructor.

=head2 c_tcp_socket( const c_tcp_socket& that )

Copy constructor - copies the other socket object,
including the file descriptor.  The two become identical,
closing one closes both, etc...

=head2 virtual ~c_tcp_socket( void )

Destroys the socket, implicitly closing the socket if
it hasn't been closed already.

=head2 int create_socket(void)

Call this to create the socket file descriptor.  
This must be called before any of the other member
functions.
Returns true on success, othewise returns false.

=head2 virtual int bind( unsigned short port )

Call this to bind the socket to a speicified port - you 
would use this member function if you were setting up a
socket to listen for a connection.
Returns true on success, othewise returns false.

=head2 virtual int connect( const char* addr, unsigned short port )

Connects a socket to a remote address/socket pair.  
Returns true on success, othewise returns false.

=head2 virtual int listen( int backlog = default_backlog )

Sets the socket up for listening on the configured port. 
You must call bind before you can call listen.
Returns true on success, othewise returns false.

=head2 virtual int accept( c_tcp_socket& client )

Call this to accept an incomming connection, once you have
bound the socket to a port, and called listen().  This call
will block untill another process has initiated a request
to your address/port pair.
Returns true on success, othewise returns false.

=head2 virtual int close( void )

Closes an open socket.
Returns true on success, othewise returns false.

=head2 virtual int send( const char*, int& )

Sends the data from the given buffer, for the number
of bytes speicfied by the int parameter.  Additionaly,
it fills the int parameter with the number of bytes
successfuly sent.
Returns true on success, othewise returns false.

=head2 virtual int send( const char* )

Used to send a C style, null termianted string.
Returns true on success, othewise returns false.

=head2 virtual int recv( char*, int& )

Used to receive data into the given buffer.  The int
parameter should specifiy the number of bytes that you
wish to receive at the
time of the method call, and will hold the number of
bytes that have been received upon return.
Returns true on success, othewise returns false.

=head2 virtual int get_line( char*, int&, char delim = '\n' )

Retreives data from the socket one character at a time untill
a delimiter is found.  The int parameter should specifiy the
size of the input buffer.  The resulting string will be null
termianted.
Returns true on success, othewise returns false.

=head2 virtual int get_line( char*, int&, char delim, char eof )

This also retreives a line of text from the socket, with the 
additional paramter of eof -- which will cause it to stop looking
for the end of a line and return if found.
Returns true on success, othewise returns false.

=head2 virtual int data_waiting(void)

Returns true if there is data waiting to be read from
the socket, othewise returns false.

=head2 virtual sock_state get_state(void)

Returns the state of the socket object.

=head2 virtual int get_socket_fd( void )

Returns the file descriptor assocaited with the socket.
Please be aware that this gives you full control over
the socket.  If you manipulate the socket file descriptor,
c_socket's behavior after that point is undefined.

=head2 virtual int operator!(void)

Used to test the status of the socket object.
Returns true unless the object is in an error state, othewise returns false.

=head2 virtual int getpeeraddr( char* buff, int len )

sprintf's the ip address of the peer into the given buffer.
Len should speicify the size of given buffer.
Returns true on success, othewise returns false.

=head2 virtual int getsockaddr( char* buff, int len )

Fills buff with the address of the socket object.
Len should speicify the size of given buffer.
Returns true on success, othewise returns false.

=head2 unsigned long name_resolve( const char* host )

Returns the ipaddress (in network byte order) as an unsigned 
long (32 bit).

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

Books on tcp/ip.  Try Stevens for starters.

=cut

*/

#if !defined( _C_SOCKET_H )
#define _C_SOCKET_H 

#include <sys/socket.h>
#include <netinet/in.h>

#include <stdinc.h>


// this class will handle tcp socket communications _only_
class c_tcp_socket {
public:
  enum sock_state { INVALID = -1, CREATED, BOUND, CONNECTED, LISTENING };

  // default constructor
  c_tcp_socket( ostream& o = cout );

  c_tcp_socket( const c_tcp_socket& that );

  // destructor
  virtual ~c_tcp_socket( void );

  // this gets called automaticly by the constructor...
  int create_socket(void);

  virtual int bind( unsigned short port );
  virtual int connect( const char* addr, unsigned short port );
  virtual int listen( int backlog = default_backlog );
  virtual int accept( c_tcp_socket& client );
  virtual int close( void );

  virtual int send( const char*, int& );
  virtual int send( const char* );
  virtual int recv( char*, int& );

  virtual int get_line( char*, int&, const char delim = '\n' );
  virtual int get_line( char*, int&, const char delim, const char eof );

  virtual int data_waiting(void);


  // should we add some member access functions?
  virtual sock_state get_state(void) const { return state; };
  virtual int get_socket_fd( void ) const { return socket_fd; };
  virtual int operator!(void) const { return state == INVALID; };

  virtual int getpeeraddr( char* buff, int len );
  virtual int getsockaddr( char* buff, int len );

  unsigned long name_resolve( const char* host ) const;

private:
  const static int default_backlog;

  int                family;
  int                type;
  int                protocol;
  int                socket_fd;
  struct sockaddr_in this_addr;
  struct sockaddr_in peer_addr;
  ostream&           out;
  sock_state         state;
};



#endif // !defined( _C_SOCKET_H )
