#ifdef IGNORE_ME
  #define ident "@(#) c_stack.h (/sccs/ri_util/include/s.c_stack.h) 1.4 02/24/99 14:20:44";
#endif

/*  c_stack.h
  Copyright 1997 Kyle R. Burton

This file is part of c_stack.h, an attempt at a win32 equivalent for
the system log daemon available under unix.

c_stack.h is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

c_stack.h is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with c_stack.h; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

=head1 NAME

c_stack - template based stack object

=head1 SYNOPSIS

  #include <c_stack.h>
  ...
  c_stack<int> primes;

  primes.push(4);
  primes.push(6);
  primes.push(8);
  primes.push(9);

=head1 DESCRIPTION

Simple template based stack object.  You can push/pop elements 
onto/off of the stack.  It also conforms to the c_collection
interface.

=head1 API REFERENCE

=head2 c_stack() 

Default constructor.

=head2 c_stack( const c_stack<T>& that ) 

Copy constructor.  The constructed stack will contain a copy
of all the elements in the copied stack object.

=head2 virtual ~c_stack()

Destructor.  Frees all memory allocated by the stack object.

=head2 virtual const c_stack<T>& operator=( const c_stack& right )

Assignment operator.  Makes the stack on the left a copy of the stack
on the right.

=head2 virtual void push( const T& elm )

Pushes a new element onto the top of the stack.

=head2 virtual int pop( T& elm )

Pops off the head of the stack, and set's elm equal to it.

=head2 virtual int get_at( int pos, T& elm ) const

Retreives the element at the specified position without
removing it from the stack.  Since the stack is a one
way data structre, it must traverse the stack node by
node to find the given position.

=head2 virtual int find_index( const T& fnd ) const

Finds the integer index of the first element in the stack
for which T::operator== returns true for the given argument.
Returns the index on success, or -1 on failure.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle. R. Burton mortis@voicenet.com

=head1 SEE ALSO

L<c_collection>

Books on data structures, try Knuth, or Sedgwick.

=cut

*/


#if !defined( _C_STACK )
#define _C_STACK

#include "c_collection.h"

template<class T>
class c_stack_node 
{
public:
  c_stack_node() 
    : m_elm() 
    , m_next( NULL ) {
  };

  c_stack_node( const T& elm )
    : m_elm( elm ) 
    , m_next( NULL ) {
  };

  c_stack_node( const c_stack_node& right )
    : m_elm( right.m_elm )
    , m_next( NULL ) {
  };

  ~c_stack_node() {
    ;// do nothing...
  };

  const c_stack_node& operator=( const c_stack_node& right ) {
    m_elm = right.m_elm;

    return( *this );
  };

public:
  T          m_elm;
  c_stack_node<T>*  m_next;
};

//
// class:  c_stack
//
// description:  template based stack class.  
//
//
template<class T>
class c_stack
  : virtual public c_collection<T>
{
public:
  // 
  // c_stack()
  //
  // default constructor
  //
  c_stack() 
    : m_top( NULL )
    , m_elms( 0 ) {
  };

  c_stack( const c_stack<T>& that ) 
    : m_top( NULL )
    , m_elms( 0 ) {
    operator=(that);
  };

  //
  // ~c_stack()
  //
  // destructor
  //
  virtual ~c_stack() {
    remove_all();
  };

  // implementation for c_collection interface
  virtual int get_count( void ) const { return( m_elms ); };

  virtual int add( const T& elm ) {
    push( elm );
    return( TRUE );
  };

  virtual const c_stack<T>& operator=( const c_stack& right ) {
    remove_all();
    c_collection_iter<T> *iter = right.prepare_iter( pre_order );
    T elm;
    while( right.next( iter, elm ) ) {
      add( elm );
    }
    right.complete_iter( iter );

    return(*this);
  };

  virtual int remove_all( void ) {
    T tmp;
    while( pop(tmp) ) ;// do nothing...
    
    return( TRUE );
  };
  virtual c_collection_iter<T>* prepare_iter( const collection_sorting_order ord = raw_order ) const {
    // the only sort order we support is c_collection::raw_order
    if( raw_order != ord && pre_order != ord ) {
      throw( c_exception::New( "unsupported iteration order", __FILE__, __LINE__ ) );
    }

    return( new c_collection_iter<T>( owner_id(this), (c_stack_node<T>*)NULL, ord ) );
  };

  virtual int first( c_collection_iter<T>* iter, T& elm ) const {
    if( (c_stack_node<T>*)NULL == m_top ) {
      iter->m_data = (void*)NULL;
      return( FALSE );
    }

    iter->m_data = (void*)m_top;
    c_stack_node<T> *n = (c_stack_node<T>*)iter->m_data;
    elm = n->m_elm;
    
    return( TRUE );
  };

  virtual int next( c_collection_iter<T>* iter, T& elm ) const {
    c_stack_node<T> *n = (c_stack_node<T>*)iter->m_data;
    if( NULL == iter->m_data ) {
      return( first( iter, elm ) );
    }

    if( (c_stack_node<T>*)NULL == ((c_stack_node<T>*)iter->m_data)->m_next ) {
      return( FALSE );
    }

    iter->m_data = (void*) ((c_stack_node<T>*)iter->m_data)->m_next;
    elm = ((c_stack_node<T>*)iter->m_data)->m_elm;

    return( TRUE );
  };

  virtual int prev( c_collection_iter<T>* iter, T& elm ) const {
    if( NULL == iter->m_data ) {
      return( last( iter, elm ) );
    }
      
    if( m_top == (c_stack_node<T>*)iter->m_data ) {
      return( FALSE );
    }

    c_stack_node<T> *tmp = m_top;
    int rv = FALSE;
    while( NULL != tmp ) {
      if( tmp->m_next == (c_stack_node<T>*)iter->m_data ) {
        iter->m_data = (void*)tmp;
        elm = ((c_stack_node<T>*)iter->m_data)->m_elm;
        rv = TRUE;
        break;
      }
      tmp = tmp->m_next;
    }

    if( !rv ) {
      iter->m_data = NULL;
    }
    return( rv );
  };

  virtual int last( c_collection_iter<T>* iter, T& elm ) const {
    if( NULL == m_top ) {
      return( FALSE );
    }

    c_stack_node<T> *tmp = m_top;
    while( NULL != tmp ) {
      tmp = tmp->m_next;
    }

    iter->m_data = (void*)tmp;
    elm = ((c_stack_node<T>*)iter->m_data)->m_elm;
    return( TRUE );
  };

  //
  // virtual int push( T& elm )
  //
  // makes a copy of elm, and pushes it onto the top of the stack
  // 
  virtual void push( const T& elm ) {
    c_stack_node<T>* n = new c_stack_node<T>( elm );
    n->m_next = m_top;
    m_top = n;
    m_elms++;
  };

  //
  // virtual int pop( T& elm )
  //
  // removes the top element, and sets elm equal to it, then deletes
  // the element.  Returns TRUE on success, FALSE on failure.
  //
  virtual int pop( T& elm ) {
    if( (c_stack_node<T>*)NULL == m_top ) {
      return( FALSE );
    }

    c_stack_node<T>* n = m_top;
    m_top = m_top->m_next;
    elm = n->m_elm;
    delete n;
    m_elms--;

    if( !m_elms ) {
      m_top = NULL;
    }

    return( TRUE );
  };

  //
  // virtual int get_at( int pos, T& elm ) const
  //
  // sets elm equal to element number pos in the stack
  // returns TRUE on success, FALSE on failure.
  //
  virtual int get_at( int pos, T& elm ) const {
    c_stack_node<T>* tmp = find( pos );

    if( (c_stack_node<T>*)NULL == tmp ) {
      return( FALSE );
    }

    elm = tmp->m_elm;
    return( TRUE );
  };

  //
  // virtual int find_index( const T& fnd ) const
  //
  // returns the index for the item that matches fnd, or
  // -1 if not found.
  //
  virtual int find_index( const T& fnd ) const {
    int pos = -1, i;
    T tmp;

    for( i = 0; i < m_elms; i++ ) {
      get_at( i, tmp );
      if( fnd == tmp ) {
        pos = 1;
        break;
      }
    }

    return( pos );
  };


protected:
  c_stack_node<T>*  m_top;
  int    m_elms;

  c_stack_node<T>* find( int pos ) const {
    c_stack_node<T>* tmp = m_top;

    if( pos < 0 || pos > m_elms ) {
      return( (c_stack_node<T>*)NULL );
    }

    while( (c_stack_node<T>*)NULL != tmp ) {
      pos--;
      tmp = tmp->m_next;
    }

    return( tmp );
  };
};

#endif // !defined( _C_STACK )
