#ifdef IGNORE_ME
  #define ident "@(#) c_thread_safe_queue.h (/sccs/ri_util/include/s.c_thread_safe_queue.h) 1.3 02/24/99 14:20:44";
#endif

/*
 * file: c_thread_safe_queue.h
 *
 * description:  this is the beginning of a thread (thread, as in 
 *               pthread_create(), not fork()) safe queue object.  
 *               The goal here is to allocate a unique semaphore
 *               upon creation (free it upon destruction), and 
 *               perform locking/unlocking within the add/remove
 *               member functions.  For now, there will be no
 *               "thread safeness", just skeleton code so we
 *               will have an easier time of doing the "thread
 *               safeness" later...
 *
 *               c_thread_safe_queue implements a FIFO queue.
 * 
 * revision list: 
 *  Wed May 13 09:39:36 EDT 1998 KRB - initial breath of life...
 *  Wed Jun 17 09:20:30 EDT 1998 KRB - working on adding the thread-ness via
 *    semaphores...please note that this will likely make this object platform
 *    dependant...
 *  Mon Jun 22 17:33:24 EDT 1998 KRB - thread safeness has been implemented
 *    with semaphores on HP-UX
 *
 */

/*

=head1 NAME

c_thread_safe_queue - template based, thread safe, queue (FIFO) object

=head1 SYNOPSIS

  #include <c_thread_safe_queue.h>
  // the rest is too complicated for a synopsis, read on...

=head1 DESCRIPTION

This object implements a thread safe queue object.  It is 
suitable to be used as a data passing mechanism between
active threads.  It is most often used in the producer/consumer
paradime of multi threaded programming, where one or more threads
produce 'jobs', and one or more threads consume those 'jobs' and
process them.

It internaly uses c_list to implement the FIFO, and c_mutex to proctect
access to it.

=head1 API REFERENCE

=head2 c_thread_safe_queue( ostream& o = cout ) 

Constructor.

=head2 virtual ~c_thread_safe_queue(void)

Destructor.

=head2 virtual int add( const M& msg )

Adds a new job to the queue.  This method utilizes the mutex
and locks out all other method calls.
Returns true on success, otherwise it returns false.

=head2 virtual int remove( M& msg )

Removes a job from the queue.  This method utilizes the mutex
and locks out all other method calls.  It will return false
if there are no more jobs on the queue.
Returns true on success, otherwise it returns false.

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton

=head1 SEE ALSO

L<c_collection>

L<c_list>

L<c_mutex>

=cut

*/

#if !defined( _RC_THREAD_SAFE_QUEUE )
#define _RC_THREAD_SAFE_QUEUE 

#include <iostream.h>

#include <c_list.h>
#include <c_mutex.h>


template< class M >
class c_thread_safe_queue
{
public:
  /*
   * birth
   *
   */
  c_thread_safe_queue( ostream& o = cout ) 
    : queue()
    , out(o) 
    , mutex(o) {
    ;// whats this new world all about?
  };

  /* 
   * death
   *
   */
  virtual ~c_thread_safe_queue(void) {
    ;// goodbye cruel world...
    // #if defined( _DEBUG )
      out << "~c_thread_safe_queue()" << endl;
    // #endif
  };

  /******************************************************************** 
   * add( const M& msg )
   *
   * add a message to the end of the queue -- please remember that
   * it's added via a copy, either via M::M(const M&), or M::operator=(const M&)
   * so M has to support those member functions...
   *
   */
  virtual int add( const M& msg ) {
    try {
      c_mutex_lock lock(mutex);
      queue.add_tail(msg);
      return TRUE;
    }
    catch( c_exception e ) {
      ERROR(out) << "[add] exception caught during mutex locking." << endl;
      throw;
    }
  };

  /******************************************************************** 
   * remove( M& msg )
   *
   * remove the oldest message from the queue.
   *
   */
  virtual int remove( M& msg ) {
    try {
      c_mutex_lock lock(mutex);
      int rv = queue.remove_head(msg);
      return rv;
    }
    catch( c_exception e ) {
      ERROR(out) << "[remove] exception caught during mutex locking." << endl;
      throw;
    }
  };

private:
  c_list<M>  queue;
  ostream&  out;
  c_mutex  mutex;
};

#endif // !defined( _RC_THREAD_SAFE_QUEUE )

