/* 
 * file:        tree_utils.h
 *
 * description: c_multi_tree related utilities.
 *
 * revision list:
 *    Thu Dec 17 10:14:18 EST 1998 KRB - created the file
 *
 *
 */

#ifdef IGNORE_ME
  #define ident "@(#) tree_utils.h (/sccs/ri_util/include/s.tree_utils.h) 1.2 02/24/99 14:20:45";
#endif

/*

=head1 NAME

tree_utils - utility functions for manipulating, and working
with c_multi_tree objects.

=head1 SYNOPSIS

  #include <tree_utils.h>

  template<class K,class V> 
    int flip_keys_and_values( c_multi_tree_manipulator<K,V>& _handle )

  template<class K,class V>
    int set_constant_value( c_multi_tree_manipulator<K,V>& _handle, const V& val )

  template<class K,class V>
    int set_constant_key( c_multi_tree_manipulator<K,V>& _handle, const K& key )

=head1 DESCRIPTION

These funcitons are for working with c_multi_tree objects.  They are all
template based, so they can match the types assocaited with the c_multi_tree
objects they are manipulating.

=head2 flip_keys_and_values

This function will walk all the nodes in the tree from the current node 
(according to the position of the handle that is passed), and switch
the key and value objects using K::operator=(const V&), 
V::operator=(const K&), K::K(), and V::V().
You can use differnt object types for the key and the value, but they must
support assignmnet to each other for this to work.

=head2 set_constant_value

set_constant_value will traverse the tree, setting the value in each
node to the supplied value.  It does this through V::operator==(const V&).

=head2 set_constant_key

set_constant_key will traverse the tree, setting the key in each
node to the supplied key.  It does this through K::operator==(const K&).

=head1 BUGS

No known bugs.

=head1 AUTHORS

Kyle R. Burton mortis@voicenet.com

=head1 SEE ALSO

L<c_multi_tree>

=cut

*/

#include <c_multi_tree.h>

#if !defined( _TREE_UTILS_H )

/* 
 * flips the key and value in the tree -- for this to work, K and V will need
 * to meet the following criteria:
 *  - K, and V must support K::K() and V::V() respectivly
 * and, if K and V are the same type:  
 *  - support K::operator=( const K& )
 * if they're different types:
 *  - K supports K::operator=( const V& )
 *  - V supports V::operator=( const K& )
 *
 */
template<class K,class V> 
  int flip_keys_and_values( c_multi_tree_manipulator<K,V>& _handle )
{
  c_multi_tree_manipulator<K,V> handle(_handle);    K key;    V val;

  handle.get_data(key,val);
  handle.set_data(val,key);

  for( int i = 0; i < handle.num_children(); ++i ) {
    handle.down(i);
    flip_keys_and_values<K,V>(handle);
    handle.up();
  }

  return TRUE;
}

/*
 * takes a single value object, and applies it to all the node values in
 * the tree - one by one, recursivly.
 *
 */
template<class K,class V>
  int set_constant_value( c_multi_tree_manipulator<K,V>& _handle, const V& val )
{
  c_multi_tree_manipulator<K,V> handle(_handle);

  handle.set_value(val);

  for( int i = 0; i < handle.num_children(); ++i ) {
    handle.down(i);
    set_constant_value<K,V>(handle,val);
    handle.up();
  }

  return TRUE;
}

/*
 * takes a single key object, and applies it to all the node keys in
 * the tree - one by one, recursivly.
 *
 */
template<class K,class V>
  int set_constant_key( c_multi_tree_manipulator<K,V>& _handle, const K& key )
{
  c_multi_tree_manipulator<K,V> handle(_handle);

  handle.set_key(key);

  for( int i = 0; i < handle.num_children(); ++i ) {
    handle.down(i);
    set_constant_key<K,V>(handle,key);
    handle.up();
  }

  return TRUE;
}

#endif
